"""

classification_postprocessing.py

Functions for postprocessing species classification results, particularly:
    
* Smoothing results within a sequence (a sequence that looks like deer/deer/deer/elk/deer/deer
  is really just a deer)
* Smoothing results within an image (an image with 700 cows and one deer is really just 701
  cows)
  
"""

#%% Constants and imports

import json
import copy

from collections import defaultdict
from tqdm import tqdm

from megadetector.utils.ct_utils import is_list_sorted


#%% Options classes

class ClassificationSmoothingOptionsImageLevel:
    """
    Options used to parameterize smooth_classification_results_image_level()
    """

    def __init__(self):
        
        #: How many detections do we need above the classification threshold to determine a dominant category
        #: for an image?
        self.min_detections_above_threshold = 4
    
        #: Even if we have a dominant class, if a non-dominant class has at least this many classifications
        #: in an image, leave them alone.
        self.max_detections_secondary_class = 3
    
        #: If the dominant class has at least this many classifications, overwrite "other" classifications
        self.min_detections_to_overwrite_other = 2
        
        #: Names to treat as "other" categories; can't be None, but can be empty
        self.other_category_names = ['other']
    
        #: What confidence threshold should we use for assessing the dominant category in an image?
        self.classification_confidence_threshold = 0.6
    
        #: Which classifications should we even bother over-writing?
        self.classification_overwrite_threshold = 0.3
    
        #: Detection confidence threshold for things we count when determining a dominant class
        self.detection_confidence_threshold = 0.2
    
        #: Which detections should we even bother over-writing?
        self.detection_overwrite_threshold = 0.05


class ClassificationSmoothingOptionsSequenceLevel:
    """
    Options used to parameterize smooth_classification_results_sequence_level()
    """
    
    def __init__(self):
        
        #: Only process detections in this category
        self.animal_detection_category = '1'
        
        #: Treat category names on this list as "other", which can be flipped to common
        #: categories.
        self.other_category_names = set(['other'])
        
        #: These are the only classes to which we're going to switch "other" classifications.
        #:
        #: Example:
        #:
        #: ['deer','elk','cow','canid','cat','bird','bear']
        self.category_names_to_smooth_to = None
        
        #: Only switch classifications to the dominant class if we see the dominant class at least
        #: this many times
        self.min_dominant_class_classifications_above_threshold_for_class_smoothing = 5 # 2
        
        #: If we see more than this many of a class that are above threshold, don't switch those
        #: classifications to the dominant class.
        self.max_secondary_class_classifications_above_threshold_for_class_smoothing = 5
        
        #: If the ratio between a dominant class and a secondary class count is greater than this, 
        #: regardless of the secondary class count, switch those classifications (i.e., ignore
        #: max_secondary_class_classifications_above_threshold_for_class_smoothing).
        #:
        #: This may be different for different dominant classes, e.g. if we see lots of cows, they really
        #: tend to be cows.  Less so for canids, so we set a higher "override ratio" for canids.
        #:
        #: Should always include a "None" category as the default ratio.
        #:
        #: Example:
        #:
        #: {'cow':2,None:3}
        self.min_dominant_class_ratio_for_secondary_override_table = {None:3}
        
        #: If there are at least this many classifications for the dominant class in a sequence,
        #: regardless of what that class is, convert all 'other' classifications (regardless of 
        #: confidence) to that class.
        self.min_dominant_class_classifications_above_threshold_for_other_smoothing = 3 # 2
        
        #: If there are at least this many classifications for the dominant class in a sequence,
        #: regardless of what that class is, classify all previously-unclassified detections
        #: as that class.
        self.min_dominant_class_classifications_above_threshold_for_unclassified_smoothing = 3 # 2
        
        #: Only count classifications above this confidence level when determining the dominant
        #: class, and when deciding whether to switch other classifications.
        self.classification_confidence_threshold = 0.6
        
        #: Confidence values to use when we change a detection's classification (the
        #: original confidence value is irrelevant at that point) (for the "other" class)
        self.flipped_other_confidence_value = 0.6
        
        #: Confidence values to use when we change a detection's classification (the
        #: original confidence value is irrelevant at that point) (for all non-other classes)
        self.flipped_class_confidence_value = 0.6
        
        #: Confidence values to use when we change a detection's classification (the
        #: original confidence value is irrelevant at that point) (for previously unclassified detections)
        self.flipped_unclassified_confidence_value = 0.6
        
        #: Only flip the class label unclassified detections if the detection confidence exceeds this threshold
        self.min_detection_confidence_for_unclassified_flipping = 0.15
        
        #: Only relevant when MegaDetector results are supplied as a dict rather than a file; determines
        #: whether smoothing happens in place.
        self.modify_in_place = True
        
# ...class ClassificationSmoothingOptionsSequenceLevel()

    
#%% Image-level smoothing

def smooth_classification_results_image_level(input_file,output_file=None,options=None):
    """
    Smooth classifications at the image level for all results in the MD-formatted results
    file [input_file], optionally writing a new set of results to [output_file].
    
    This function generally expresses the notion that an image with 700 cows and one deer 
    is really just 701 cows.
    
    Only count detections with a classification confidence threshold above
    [options.classification_confidence_threshold], which in practice means we're only
    looking at one category per detection.
    
    If an image has at least [options.min_detections_above_threshold] such detections
    in the most common category, and no more than [options.max_detections_secondary_class]
    in the second-most-common category, flip all detections to the most common
    category.
    
    Optionally treat some classes as particularly unreliable, typically used to overwrite an 
    "other" class.
    
    This function also removes everything but the non-dominant classification for each detection.
    
    Args:
        input_file (str): MegaDetector-formatted classification results file to smooth
        output_file (str, optional): .json file to write smoothed results
        options (ClassificationSmoothingOptionsImageLevel, optional): see 
          ClassificationSmoothingOptionsImageLevel for details.
            
    Returns:
        dict: MegaDetector-results-formatted dict, identical to what's written to
        [output_file] if [output_file] is not None.
    """
    
    if options is None:
        options = ClassificationSmoothingOptionsImageLevel()
        
    with open(input_file,'r') as f:
        print('Loading results from:\n{}'.format(input_file))
        d = json.load(f)
        
    category_name_to_id = {d['classification_categories'][k]:k for k in d['classification_categories']}
    other_category_ids = []
    for s in options.other_category_names:
        if s in category_name_to_id:
            other_category_ids.append(category_name_to_id[s])
        else:
            print('Warning: "other" category {} not present in file {}'.format(
                s,input_file))
    
    n_other_classifications_changed = 0
    n_other_images_changed = 0
    
    n_detections_flipped = 0
    n_images_changed = 0
    
    # Before we do anything else, get rid of everything but the top classification
    # for each detection.
    for im in tqdm(d['images']):        
        if 'detections' not in im or im['detections'] is None or len(im['detections']) == 0:
            continue
        
        detections = im['detections']
        
        for det in detections:
            
            if 'classifications' not in det or len(det['classifications']) == 0:
                continue
            
            classification_confidence_values = [c[1] for c in det['classifications']]
            assert is_list_sorted(classification_confidence_values,reverse=True)
            det['classifications'] = [det['classifications'][0]]
    
        # ...for each detection in this image
        
    # ...for each image
    
    # im = d['images'][0]    
    for im in tqdm(d['images']):
        
        if 'detections' not in im or im['detections'] is None or len(im['detections']) == 0:
            continue
        
        detections = im['detections']
    
        category_to_count = defaultdict(int)
        for det in detections:
            if ('classifications' in det) and (det['conf'] >= options.detection_confidence_threshold):
                for c in det['classifications']:
                    if c[1] >= options.classification_confidence_threshold:
                        category_to_count[c[0]] += 1
                # ...for each classification
            # ...if there are classifications for this detection
        # ...for each detection
                        
        if len(category_to_count) <= 1:
            continue
        
        category_to_count = {k: v for k, v in sorted(category_to_count.items(),
                                                     key=lambda item: item[1], 
                                                     reverse=True)}
        
        keys = list(category_to_count.keys())
        
        # Handle a quirky special case: if the most common category is "other" and 
        # it's "tied" with the second-most-common category, swap them
        if (len(keys) > 1) and \
            (keys[0] in other_category_ids) and \
            (keys[1] not in other_category_ids) and \
            (category_to_count[keys[0]] == category_to_count[keys[1]]):
                keys[1], keys[0] = keys[0], keys[1]
        
        max_count = category_to_count[keys[0]]
        # secondary_count = category_to_count[keys[1]]
        # The 'secondary count' is the most common non-other class
        secondary_count = 0
        for i_key in range(1,len(keys)):
            if keys[i_key] not in other_category_ids:
                secondary_count = category_to_count[keys[i_key]]
                break

        most_common_category = keys[0]
        
        assert max_count >= secondary_count
        
        # If we have at least *min_detections_to_overwrite_other* in a category that isn't
        # "other", change all "other" classifications to that category
        if max_count >= options.min_detections_to_overwrite_other and \
            most_common_category not in other_category_ids:
            
            other_change_made = False
            
            for det in detections:
                
                if ('classifications' in det) and \
                    (det['conf'] >= options.detection_overwrite_threshold): 
                    
                    for c in det['classifications']:                
                        
                        if c[1] >= options.classification_overwrite_threshold and \
                            c[0] in other_category_ids:
                                
                            n_other_classifications_changed += 1
                            other_change_made = True
                            c[0] = most_common_category
                            
                    # ...for each classification
                    
                # ...if there are classifications for this detection
                
            # ...for each detection
            
            if other_change_made:
                n_other_images_changed += 1
            
        # ...if we should overwrite all "other" classifications
    
        if max_count < options.min_detections_above_threshold:
            continue
        
        if secondary_count >= options.max_detections_secondary_class:
            continue
        
        # At this point, we know we have a dominant category; change all other above-threshold
        # classifications to that category.  That category may have been "other", in which
        # case we may have already made the relevant changes.
        
        n_detections_flipped_this_image = 0
        
        # det = detections[0]
        for det in detections:
            
            if ('classifications' in det) and \
                (det['conf'] >= options.detection_overwrite_threshold):
                
                for c in det['classifications']:
                    if c[1] >= options.classification_overwrite_threshold and \
                        c[0] != most_common_category:
                            
                        c[0] = most_common_category
                        n_detections_flipped += 1
                        n_detections_flipped_this_image += 1
                
                # ...for each classification
                
            # ...if there are classifications for this detection
            
        # ...for each detection
        
        if n_detections_flipped_this_image > 0:
            n_images_changed += 1
    
    # ...for each image    
    
    print('Classification smoothing: changed {} detections on {} images'.format(
        n_detections_flipped,n_images_changed))
    
    print('"Other" smoothing: changed {} detections on {} images'.format(
          n_other_classifications_changed,n_other_images_changed))

    if output_file is not None:    
        print('Writing results after image-level smoothing to:\n{}'.format(output_file))
        with open(output_file,'w') as f:
            json.dump(d,f,indent=1)        

    return d

# ...def smooth_classification_results_image_level(...)


#%% Sequence-level smoothing

def _results_for_sequence(images_this_sequence,filename_to_results):
    """
    Fetch MD results for every image in this sequence, based on the 'file_name' field
    """
    
    results_this_sequence = []
    for im in images_this_sequence:
        fn = im['file_name']
        results_this_image = filename_to_results[fn]
        assert isinstance(results_this_image,dict)
        results_this_sequence.append(results_this_image)
        
    return results_this_sequence
            
    
def _top_classifications_for_sequence(images_this_sequence,filename_to_results,options):
    """
    Return all top-1 animal classifications for every detection in this 
    sequence, regardless of  confidence

    May modify [images_this_sequence] (removing non-top-1 classifications)
    """
    
    classifications_this_sequence = []

    # im = images_this_sequence[0]
    for im in images_this_sequence:
        
        fn = im['file_name']
        results_this_image = filename_to_results[fn]
        
        if results_this_image['detections'] is None:
            continue
        
        # det = results_this_image['detections'][0]
        for det in results_this_image['detections']:
            
            # Only process animal detections
            if det['category'] != options.animal_detection_category:
                continue
            
            # Only process detections with classification information
            if 'classifications' not in det:
                continue
            
            # We only care about top-1 classifications, remove everything else
            if len(det['classifications']) > 1:
                
                # Make sure the list of classifications is already sorted by confidence
                classification_confidence_values = [c[1] for c in det['classifications']]
                assert is_list_sorted(classification_confidence_values,reverse=True)
                
                # ...and just keep the first one
                det['classifications'] = [det['classifications'][0]]
                
            # Confidence values should be sorted within a detection; verify this, and ignore 
            top_classification = det['classifications'][0]
            
            classifications_this_sequence.append(top_classification)
    
        # ...for each detection in this image
        
    # ...for each image in this sequence

    return classifications_this_sequence

# ..._top_classifications_for_sequence()


def _count_above_threshold_classifications(classifications_this_sequence,options):
    """
    Given a list of classification objects (tuples), return a dict mapping
    category IDs to the count of above-threshold classifications.
    
    This dict's keys will be sorted in descending order by frequency.
    """
    
    # Count above-threshold classifications in this sequence
    category_to_count = defaultdict(int)
    for c in classifications_this_sequence:
        if c[1] >= options.classification_confidence_threshold:
            category_to_count[c[0]] += 1
    
    # Sort the dictionary in descending order by count
    category_to_count = {k: v for k, v in sorted(category_to_count.items(),
                                                 key=lambda item: item[1], 
                                                 reverse=True)}
    
    keys_sorted_by_frequency = list(category_to_count.keys())
        
    # Handle a quirky special case: if the most common category is "other" and 
    # it's "tied" with the second-most-common category, swap them.
    if (options.other_category_names is not None) and (len(options.other_category_names) > 0):
        if (len(keys_sorted_by_frequency) > 1) and \
            (keys_sorted_by_frequency[0] in options.other_category_names) and \
            (keys_sorted_by_frequency[1] not in options.other_category_names) and \
            (category_to_count[keys_sorted_by_frequency[0]] == \
             category_to_count[keys_sorted_by_frequency[1]]):
                keys_sorted_by_frequency[1], keys_sorted_by_frequency[0] = \
                    keys_sorted_by_frequency[0], keys_sorted_by_frequency[1]

    sorted_category_to_count = {}    
    for k in keys_sorted_by_frequency:
        sorted_category_to_count[k] = category_to_count[k]
        
    return sorted_category_to_count

# ...def _count_above_threshold_classifications()
 
   
def _sort_images_by_time(images):
    """
    Returns a copy of [images], sorted by the 'datetime' field (ascending).
    """
    return sorted(images, key = lambda im: im['datetime'])        
    

def _get_first_key_from_sorted_dictionary(di):
    if len(di) == 0:
        return None
    return next(iter(di.items()))[0]


def _get_first_value_from_sorted_dictionary(di):
    if len(di) == 0:
        return None
    return next(iter(di.items()))[1]


def smooth_classification_results_sequence_level(md_results,
                                                 cct_sequence_information,
                                                 output_file=None,
                                                 options=None):
    """
    Smooth classifications at the sequence level for all results in the MD-formatted results
    file [md_results_file], optionally writing a new set of results to [output_file].
    
    This function generally expresses the notion that a sequence that looks like
    deer/deer/deer/elk/deer/deer/deer/deer is really just a deer.
    
    Args:
        md_results (str or dict): MegaDetector-formatted classification results file to smooth
          (or already-loaded results).  If you supply a dict, it's modified in place by default, but
          a copy can be forced by setting options.modify_in_place=False.
        cct_sequence_information (str, dict, or list): COCO Camera Traps file containing sequence IDs for
          each image (or an already-loaded CCT-formatted dict, or just the 'images' list from a CCT dict).
        output_file (str, optional): .json file to write smoothed results
        options (ClassificationSmoothingOptionsSequenceLevel, optional): see 
          ClassificationSmoothingOptionsSequenceLevel for details.
            
    Returns:
        dict: MegaDetector-results-formatted dict, identical to what's written to
        [output_file] if [output_file] is not None.
    """
    
    if options is None:
        options = ClassificationSmoothingOptionsSequenceLevel()
    
    if options.category_names_to_smooth_to is None:
        options.category_names_to_smooth_to = []
    
    if options.other_category_names is None:
        options.other_category_names = []
        
    assert None in options.min_dominant_class_ratio_for_secondary_override_table, \
        'Oops, it looks like you removed the default (None) key from ' + \
            'options.min_dominant_class_ratio_for_secondary_override_table'
    
    if isinstance(md_results,str):
        print('Loading MD results from {}'.format(md_results))
        with open(md_results,'r') as f:
            md_results = json.load(f)
    else:
        assert isinstance(md_results,dict)
        if not options.modify_in_place:
            print('Copying MD results instead of modifying in place')
            md_results = copy.deepcopy(md_results)
        else:
            print('Smoothing MD results in place')
    
    if isinstance(cct_sequence_information,list):
        image_info = cct_sequence_information
    elif isinstance(cct_sequence_information,str):
        print('Loading sequence information from {}'.format(cct_sequence_information))
        with open(cct_sequence_information,'r') as f:
            cct_sequence_information = json.load(f)
            image_info = cct_sequence_information['images']
    else:
        assert isinstance(cct_sequence_information,dict)
        image_info = cct_sequence_information['images']
        
    
    ##%% Make a list of images appearing at each location
    
    sequence_to_images = defaultdict(list)
    
    # im = image_info[0]
    for im in tqdm(image_info):
        sequence_to_images[im['seq_id']].append(im)
    
    all_sequences = list(sorted(sequence_to_images.keys()))
    
    
    ##%% Load classification results
    
    # Map each filename to classification results for that file
    filename_to_results = {}
    
    for im in tqdm(md_results['images']):
        filename_to_results[im['file'].replace('\\','/')] = im
    
    
    ##%% Smooth classification results over sequences (prep)
    
    classification_category_id_to_name = md_results['classification_categories']
    classification_category_name_to_id = {v: k for k, v in classification_category_id_to_name.items()}
    
    class_names = list(classification_category_id_to_name.values())
    
    assert(md_results['detection_categories'][options.animal_detection_category] == 'animal')
    
    other_category_ids = set([classification_category_name_to_id[s] for s in options.other_category_names])
    
    category_ids_to_smooth_to = set([classification_category_name_to_id[s] for s in options.category_names_to_smooth_to])
    assert all([s in class_names for s in options.category_names_to_smooth_to])    
    
    
    ##%% Smooth classifications at the sequence level (main loop)
    
    n_other_flips = 0
    n_classification_flips = 0
    n_unclassified_flips = 0
    
    # Break if this token is contained in a filename (set to None for normal operation)
    debug_fn = None
    
    # i_sequence = 0; seq_id = all_sequences[i_sequence]
    for i_sequence,seq_id in tqdm(enumerate(all_sequences),total=len(all_sequences)):
        
        images_this_sequence = sequence_to_images[seq_id]
        
        # Count top-1 classifications in this sequence (regardless of confidence)
        classifications_this_sequence = _top_classifications_for_sequence(images_this_sequence,
                                                                          filename_to_results,
                                                                          options)
        
        # Handy debugging code for looking at the numbers for a particular sequence
        for im in images_this_sequence:
            if debug_fn is not None and debug_fn in im['file_name']:
                raise ValueError('')
                 
        if len(classifications_this_sequence) == 0:
            continue
        
        # Count above-threshold classifications for each category
        sorted_category_to_count = _count_above_threshold_classifications(
            classifications_this_sequence,options)
        
        if len(sorted_category_to_count) == 0:
            continue
        
        max_count = _get_first_value_from_sorted_dictionary(sorted_category_to_count)    
        dominant_category_id = _get_first_key_from_sorted_dictionary(sorted_category_to_count)
        
        # If our dominant category ID isn't something we want to smooth to, 
        # don't mess around with this sequence
        if dominant_category_id not in category_ids_to_smooth_to:
            continue
            
        
        ## Smooth "other" classifications ##
        
        if max_count >= options.min_dominant_class_classifications_above_threshold_for_other_smoothing:        
            for c in classifications_this_sequence:           
                if c[0] in other_category_ids:
                    n_other_flips += 1
                    c[0] = dominant_category_id
                    c[1] = options.flipped_other_confidence_value
    
    
        # By not re-computing "max_count" here, we are making a decision that the count used
        # to decide whether a class should overwrite another class does not include any "other"
        # classifications we changed to be the dominant class.  If we wanted to include those...
        # 
        # sorted_category_to_count = count_above_threshold_classifications(classifications_this_sequence)
        # max_count = get_first_value_from_sorted_dictionary(sorted_category_to_count)    
        # assert dominant_category_id == get_first_key_from_sorted_dictionary(sorted_category_to_count)
        
        
        ## Smooth non-dominant classes ##
        
        if max_count >= options.min_dominant_class_classifications_above_threshold_for_class_smoothing:
            
            # Don't flip classes to the dominant class if they have a large number of classifications
            category_ids_not_to_flip = set()
            
            for category_id in sorted_category_to_count.keys():
                secondary_class_count = sorted_category_to_count[category_id]
                dominant_to_secondary_ratio = max_count / secondary_class_count
                
                # Don't smooth over this class if there are a bunch of them, and the ratio
                # if primary to secondary class count isn't too large
                
                # Default ratio
                ratio_for_override = options.min_dominant_class_ratio_for_secondary_override_table[None]
                
                # Does this dominant class have a custom ratio?
                dominant_category_name = classification_category_id_to_name[dominant_category_id]
                if dominant_category_name in options.min_dominant_class_ratio_for_secondary_override_table:
                    ratio_for_override = \
                        options.min_dominant_class_ratio_for_secondary_override_table[dominant_category_name]
                        
                if (dominant_to_secondary_ratio < ratio_for_override) and \
                    (secondary_class_count > \
                     options.max_secondary_class_classifications_above_threshold_for_class_smoothing):
                    category_ids_not_to_flip.add(category_id)
                    
            for c in classifications_this_sequence:
                if c[0] not in category_ids_not_to_flip and c[0] != dominant_category_id:
                    c[0] = dominant_category_id
                    c[1] = options.flipped_class_confidence_value
                    n_classification_flips += 1
            
            
        ## Smooth unclassified detections ##
            
        if max_count >= options.min_dominant_class_classifications_above_threshold_for_unclassified_smoothing:
            
            results_this_sequence = _results_for_sequence(images_this_sequence,filename_to_results)
            detections_this_sequence = []
            for r in results_this_sequence:
                if r['detections'] is not None:
                    detections_this_sequence.extend(r['detections'])
            for det in detections_this_sequence:
                if 'classifications' in det and len(det['classifications']) > 0:
                    continue
                if det['category'] != options.animal_detection_category:
                    continue
                if det['conf'] < options.min_detection_confidence_for_unclassified_flipping:
                    continue
                det['classifications'] = [[dominant_category_id,options.flipped_unclassified_confidence_value]]
                n_unclassified_flips += 1
                                
    # ...for each sequence    
        
    print('\Finished sequence smoothing\n')
    print('Flipped {} "other" classifications'.format(n_other_flips))
    print('Flipped {} species classifications'.format(n_classification_flips))
    print('Flipped {} unclassified detections'.format(n_unclassified_flips))
        
    
    ##%% Write smoothed classification results
    
    if output_file is not None:
        
        print('Writing sequence-smoothed classification results to {}'.format(
            output_file))
        
        with open(output_file,'w') as f:
            json.dump(md_results,f,indent=1)
            
    return md_results

# ...smooth_classification_results_sequence_level(...)


"""

tf_detector.py

Module containing the class TFDetector, for loading and running a TensorFlow detection model.

"""

#%% Imports and constants

import numpy as np

from megadetector.detection.run_detector import \
    CONF_DIGITS, COORD_DIGITS, FAILURE_INFER
from megadetector.utils.ct_utils import truncate_float

import tensorflow.compat.v1 as tf

print('TensorFlow version:', tf.__version__)
print('Is GPU available? tf.test.is_gpu_available:', tf.test.is_gpu_available())


#%% Classes

class TFDetector:
    """
    A detector model loaded at the time of initialization. It is intended to be used with
    TensorFlow-based versions of MegaDetector (v2, v3, or v4).  If someone can find v1, I 
    suppose you could use this class for v1 also.
    """
    
    #: TF versions of MD were trained with batch size of 1, and the resizing function is a 
    #: part of the inference graph, so this is fixed.
    #:
    #: :meta private:  
    BATCH_SIZE = 1


    def __init__(self, model_path):
        """
        Loads a model from [model_path] and starts a tf.Session with this graph. Obtains
        input and output tensor handles.
        """
        
        detection_graph = TFDetector.__load_model(model_path)
        self.tf_session = tf.Session(graph=detection_graph)
        self.image_tensor = detection_graph.get_tensor_by_name('image_tensor:0')
        self.box_tensor = detection_graph.get_tensor_by_name('detection_boxes:0')
        self.score_tensor = detection_graph.get_tensor_by_name('detection_scores:0')
        self.class_tensor = detection_graph.get_tensor_by_name('detection_classes:0')


    @staticmethod
    def __round_and_make_float(d, precision=4):
        return truncate_float(float(d), precision=precision)


    @staticmethod
    def __convert_coords(tf_coords):
        """
        Converts coordinates from the model's output format [y1, x1, y2, x2] to the
        format used by our API and MegaDB: [x1, y1, width, height]. All coordinates
        (including model outputs) are normalized in the range [0, 1].

        Args:
            tf_coords: np.array of predicted bounding box coordinates from the TF detector,
                has format [y1, x1, y2, x2]

        Returns: list of Python float, predicted bounding box coordinates [x1, y1, width, height]
        """
        
        # change from [y1, x1, y2, x2] to [x1, y1, width, height]
        width = tf_coords[3] - tf_coords[1]
        height = tf_coords[2] - tf_coords[0]

        new = [tf_coords[1], tf_coords[0], width, height]  # must be a list instead of np.array

        # convert numpy floats to Python floats
        for i, d in enumerate(new):
            new[i] = TFDetector.__round_and_make_float(d, precision=COORD_DIGITS)
        return new


    @staticmethod
    def __load_model(model_path):
        """
        Loads a detection model (i.e., create a graph) from a .pb file.

        Args:
            model_path: .pb file of the model.

        Returns: the loaded graph.
        """
        
        print('TFDetector: Loading graph...')
        detection_graph = tf.Graph()
        with detection_graph.as_default():
            od_graph_def = tf.GraphDef()
            with tf.gfile.GFile(model_path, 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')
        print('TFDetector: Detection graph loaded.')

        return detection_graph


    def _generate_detections_one_image(self, image):
        """
        Runs the detector on a single image.
        """
        
        np_im = np.asarray(image, np.uint8)
        im_w_batch_dim = np.expand_dims(np_im, axis=0)

        # need to change the above line to the following if supporting a batch size > 1 and resizing to the same size
        # np_images = [np.asarray(image, np.uint8) for image in images]
        # images_stacked = np.stack(np_images, axis=0) if len(images) > 1 else np.expand_dims(np_images[0], axis=0)

        # performs inference
        (box_tensor_out, score_tensor_out, class_tensor_out) = self.tf_session.run(
            [self.box_tensor, self.score_tensor, self.class_tensor],
            feed_dict={self.image_tensor: im_w_batch_dim})

        return box_tensor_out, score_tensor_out, class_tensor_out


    def generate_detections_one_image(self, 
                                      image, 
                                      image_id, 
                                      detection_threshold, 
                                      image_size=None,
                                      skip_image_resizing=False,
                                      augment=False):
        """
        Runs the detector on an image.

        Args:
            image (Image): the PIL Image object on which we should run the detector
            image_id (str): a path to identify the image; will be in the "file" field of the output object            
            detection_threshold (float): only detections above this threshold will be included in the return
                value
            image_size (tuple, optional): image size to use for inference, only mess with this
                if (a) you're using a model other than MegaDetector or (b) you know what you're
                doing
            skip_image_resizing (bool, optional): whether to skip internal image resizing (and rely on external 
                resizing).  Not currently supported, but included here for compatibility with PTDetector.
            augment (bool, optional): enable image augmentation.  Not currently  supported, but included 
                here for compatibility with PTDetector.

        Returns:
            dict: a dictionary with the following fields:
                - 'file' (filename, always present)
                - 'max_detection_conf' (removed from MegaDetector output files by default, but generated here)
                - 'detections' (a list of detection objects containing keys 'category', 'conf', and 'bbox')
                - 'failure' (a failure string, or None if everything went fine)
        """
        
        assert image_size is None, 'Image sizing not supported for TF detectors'
        assert not skip_image_resizing, 'Image sizing not supported for TF detectors'
        assert not augment, 'Image augmentation is not supported for TF detectors'
        
        if detection_threshold is None:
            detection_threshold = 0
            
        result = { 'file': image_id }
        
        try:
            b_box, b_score, b_class = self._generate_detections_one_image(image)

            # our batch size is 1; need to loop the batch dim if supporting batch size > 1
            boxes, scores, classes = b_box[0], b_score[0], b_class[0]

            detections_cur_image = []  # will be empty for an image with no confident detections
            max_detection_conf = 0.0
            for b, s, c in zip(boxes, scores, classes):
                if s > detection_threshold:
                    detection_entry = {
                        'category': str(int(c)),  # use string type for the numerical class label, not int
                        'conf': truncate_float(float(s),  # cast to float for json serialization
                                               precision=CONF_DIGITS),
                        'bbox': TFDetector.__convert_coords(b)
                    }
                    detections_cur_image.append(detection_entry)
                    if s > max_detection_conf:
                        max_detection_conf = s

            result['max_detection_conf'] = truncate_float(float(max_detection_conf),
                                                          precision=CONF_DIGITS)
            result['detections'] = detections_cur_image

        except Exception as e:
            result['failure'] = FAILURE_INFER
            print('TFDetector: image {} failed during inference: {}'.format(image_id, str(e)))

        return result

    # ...def generate_detections_one_image(...)
    
# ...class TFDetector



'''
copy_and_unzip_emammal.py

Siyu Yang

Script to copy all deployments in the emammal container (mounted on the VM or not) to data
disk at /datadrive and unzip them, deleting the copied zip file.

Need to add exception handling.
'''

#%% Imports and constants

from datetime import datetime
import itertools
import json
import multiprocessing
from multiprocessing.dummy import Pool as ThreadPool  # this functions like threading
import os
from shutil import copy, copyfile
from tqdm import tqdm
from typing import Optional
import zipfile

from azure.storage.blob import BlobServiceClient


# configurations and paths
log_folder = '/home/lynx/logs'
dest_folder = '/datadrive/emammal_robertlong'  # data disk attached where data is stored
origin = 'cloud'  # 'cloud' or 'mounted'


#%% Helper functions

def _copy_unzip(source_path, dest_folder):
    
    try:
        dest_subfolder = os.path.join(dest_folder, os.path.basename(source_path).split('.zip')[0])
        if os.path.exists(dest_subfolder):
            print('{} exists.'.format(dest_subfolder))
            return('exists')

        print('Copying...')
        # dest_path = copy(source_path, dest_folder)
        dest_path = os.path.join(dest_folder, os.path.basename(source_path))
        copyfile(source_path, dest_path)

        with zipfile.ZipFile(dest_path, 'r') as zip_ref:
            zip_ref.extractall(dest_subfolder)

        os.remove(dest_path)
        print('{} copied and extracted'.format(dest_subfolder))
        return None

    except Exception:
        try:
            print('Retrying...')
            dest_path = os.path.join(dest_folder, os.path.basename(source_path))
            copyfile(source_path, dest_path)
            with zipfile.ZipFile(dest_path, 'r') as zip_ref:
                zip_ref.extractall(dest_subfolder)
            os.remove(dest_path)
            print('{} copied and extracted'.format(dest_subfolder))
            return (None)
        except Exception as e:
            print('WARNING {} did not get copied. Exception: {}'.format(source_path, str(e)))
            return source_path


def copy_from_mounted_container(source_folder, dest_folder):
    
    sources = []

    collections = sorted(os.listdir(source_folder))

    for collection in collections:
        collection_folder = os.path.join(source_folder, collection)
        if not os.path.isdir(collection_folder):
            continue

        print('Processing collection {}'.format(collection))

        for file in tqdm(sorted(os.listdir(collection_folder))):
            source_path = os.path.join(collection_folder, file)
            sources.append(source_path)

    # num_workers = multiprocessing.cpu_count()
    # pool = ThreadPool(num_workers)
    # results = pool.starmap(_copy_unzip, zip(sources, itertools.repeat(dest_folder)))
    #
    # print('Waiting for processes to finish...')
    # pool.close()
    # pool.join()

    # sequential
    results = []
    for source_path in sources:
        result = _copy_unzip(source_path, dest_folder)
        results.append(result)

    cur_time = datetime.now().strftime('%Y%m%d-%H%M%S')
    with open(os.path.join(log_folder, 'copy_unzip_results_{}.json'.format(cur_time)), 'w') as f:
        json.dump(results, f)


def _download_unzip(blob_service: BlobServiceClient,
                    container: str,
                    blob_path: str,
                    dest_path: str) -> Optional[str]:
    try:
        with open(dest_path, 'wb') as f:
            cc = blob_service.get_container_client(container)
            print('Downloading...')
            cc.download_blob(blob_path).readinto(f)

        dest_subfolder = dest_path.split('.zip')[0]

        with zipfile.ZipFile(dest_path, 'r') as zip_ref:
            zip_ref.extractall(dest_subfolder)

        os.remove(dest_path)
        print('{} copied and extracted'.format(dest_subfolder))
        return None

    except Exception as e:
        print('ERROR while downloading or unzipping {}. Exception: {}'.format(blob_path, str(e)))
        return blob_path


def download_from_container(dest_folder: str,
                            blob_service: BlobServiceClient,
                            container: str = 'emammal',
                            desired_blob_prefix: str = '') -> None:
    generator = blob_service.get_containre_client(container).list_blobs()
    desired_blobs = [blob.name for blob in generator
                     if blob.name.startswith(desired_blob_prefix)]

    print('desired_blobs', desired_blobs)

    results = []
    for blob_path in tqdm(desired_blobs):
        blob_name = blob_path.split('/')[2]
        print('blob_name', blob_name)
        dest_path = os.path.join(dest_folder, blob_name)
        print('dest_path', dest_path)
        result = _download_unzip(blob_service, container, blob_path, dest_path)
        results.append(result)

    cur_time = datetime.now().strftime('%Y%m%d-%H%M%S')
    with open(os.path.join(log_folder, 'download_unzip_results_{}.json'.format(cur_time)), 'w') as f:
        json.dump(results, f)


#%% Command-line driver
        
if __name__ == '__main__':
    
    if origin == 'cloud':
        container = 'wpz'
        desired_blob_prefix = 'emammal/0Robert Long/'

    print('Start timing...')
    start_time = datetime.now()

    if origin == 'mounted':
        # if the blob container is already mounted on the VM
        source_folder = '/home/yasiyu/mnt/wildlifeblobssc/emammal'  # blob container mounted
        copy_from_mounted_container(source_folder, dest_folder)

    elif origin == 'cloud':
        # or you can download them using the storage Python SDK
        # store storage account key in environment variable AZ_STORAGE_KEY
        blob_service = BlobServiceClient(
            account_url='wildlifeblobssc.blob.core.windows.net',
            credential=os.environ["AZ_STORAGE_KEY"])
        download_from_container(dest_folder, blob_service, container=container,
                                desired_blob_prefix=desired_blob_prefix)

    print('Process took {}.'.format(datetime.now() - start_time))



"""

 jb_csv_to_json.py

 Convert a particular .csv file to CCT format.  Images were not available at
 the time I wrote this script, so this is much shorter than other scripts 
 in this folder.

"""

#%% Constants and environment

import pandas as pd
import uuid
import json

input_metadata_file = r'd:\temp\pre_bounding_box.csv'
output_file = r'd:\temp\pre_bounding_box.json'
filename_col = 'filename'
label_col = 'category'


#%% Read source data

input_metadata = pd.read_csv(input_metadata_file)

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))


#%% Confirm filename uniqueness (this data set has one label per image)

imageFilenames = input_metadata[filename_col]

duplicateRows = []
filenamesToRows = {}

# Build up a map from filenames to a list of rows, checking image existence as we go
for iFile,fn in enumerate(imageFilenames):
    
    if (fn in filenamesToRows):
        duplicateRows.append(iFile)
        filenamesToRows[fn].append(iFile)
    else:
        filenamesToRows[fn] = [iFile]

assert(len(duplicateRows) == 0)    
    

#%% Create CCT dictionaries

images = []
annotations = []

# Map categories to integer IDs (that's what COCO likes)
nextCategoryID = 1
categories = []
categoryNamesToCategories = {}

cat = {}
cat['name'] = 'empty'
cat['id'] = 0
categories.append(cat)    
categoryNamesToCategories['empty'] = cat

# For each image
#
# Because in practice images are 1:1 with annotations in this data set,
# this is also a loop over annotations.

# imageName = imageFilenames[0]
for imageName in imageFilenames:
    
    rows = filenamesToRows[imageName]
    
    # As per above, this is convenient and appears to be true; asserting to be safe
    assert(len(rows) == 1)    
    iRow = rows[0]
    
    row = input_metadata.iloc[iRow]
    
    im = {}
    # Filenames look like "290716114012001a1116.jpg"
    im['id'] = imageName.split('.')[0]
    im['file_name'] = imageName    
    im['seq_id'] = '-1'
    
    images.append(im)
    
    categoryName = row[label_col].lower()
    
    # Have we seen this category before?
    if categoryName in categoryNamesToCategories:
        categoryID = categoryNamesToCategories[categoryName]['id']
    else:
        cat = {}
        categoryID = nextCategoryID 
        cat['name'] = categoryName
        cat['id'] = nextCategoryID 
        categories.append(cat)    
        categoryNamesToCategories[categoryName] = cat
        nextCategoryID += 1
        
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = categoryID
    
    annotations.append(ann)
    
# ...for each image
        
print('Finished creating dictionaries')


#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'COCO style database'
info['secondary_contributor'] = 'Converted to COCO .json by Dan Morris'
info['contributor'] = ''


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_file,'w'), indent=4)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
sortedCategories,data = integrity_check_json_db.integrity_check_json_db(output_file, options)





"""

 add_nacti_sizes.py

 NACTI bounding box metadata was posted before we inclduded width and height as semi-standard
 fields; pull size information from the main metadata file and add to the bbox file.

"""

#%% Constants and environment

import json
from tqdm import tqdm

input_file = 'G:/temp/nacti_metadata.json'
input_bbox_file = 'G:/temp/nacti_20200401_bboxes.json'
output_bbox_file = 'G:/temp/nacti_20230920_bboxes.json'


#%% Read .json files

with open(input_file,'r') as f:
    input_data = json.load(f)

with open(input_bbox_file,'r') as f:
    input_bbox_data = json.load(f)
    
print('Finished reading .json data')


#%% Map image names to width and height

filename_to_size = {}
for im in tqdm(input_data['images']):
    filename_to_size[im['file_name']] = (im['width'],im['height'])
    

#%% Add to output data

for im in tqdm(input_bbox_data['images']):
    size = filename_to_size[im['file_name']]
    im['width'] = size[0]
    im['height'] = size[1]
    

#%% Write output

output_bbox_data = input_bbox_data
output_bbox_data['version'] = '2023-09-20'

with open(output_bbox_file,'w') as f:
    json.dump(output_bbox_data,f,indent=1)


"""

 helena_to_cct.py

 Convert the Helena Detections data set to a COCO-camera-traps .json file

"""

#%% Constants and environment

import os
import json
import uuid
import time
import humanfriendly
import numpy as np

import pandas as pd

from PIL import Image
from datetime import datetime

from megadetector.utils.path_utils import find_images

base_directory = r'/mnt/blobfuse/wildlifeblobssc/'
output_directory = r'/home/gramener'
output_json_file = os.path.join(output_directory,'rspb.json')
input_metadata_file = os.path.join(base_directory, 'StHelena_Detections.xlsx')
image_directory = os.path.join(base_directory, 'StHELENA_images/')
mapping_df = ''
filename_col = 'image_name'
load_width_and_height = True
annotation_fields_to_copy = ['Fortnight', 'Detector', 'datetime', 'site']

assert(os.path.isdir(image_directory))

# This is one time process
#%% Create Filenames and timestamps mapping CSV

image_full_paths = find_images(image_directory, bRecursive=True)
csv_file = os.path.join(output_directory, "mapping_names.csv")
if not os.path.exists(csv_file):
    map_list = []
    for img_ in image_full_paths:
        try:
            date_cr = Image.open(img_)._getexif()[306]
            _tmp = {}
            # import pdb;pdb.set_trace()
            img_path = img_.replace(image_directory, "")
            img_folder = img_path.split("/")[0]
            site = img_path.split("/")[1]
            detector = img_path.split("/")[2]
            _tmp["image_name"] = img_path
            _tmp["Fortnight"] = img_folder.replace("Fortnight", "")
            _tmp["site"] = site
            _tmp["Detector"] = detector
            _tmp["datetime"] = "-".join(date_cr.split(":")[:-1])
            map_list.append(_tmp)
        except Exception as e:
            print(e)
            print(img_)
    mapping_df = pd.DataFrame(map_list)
    mapping_df.to_csv(csv_file, index=False)
else:
    mapping_df = pd.read_csv(csv_file)

#%% To create CCT JSON for RSPB dataset

#%% Read source data
input_metadata = pd.read_excel(input_metadata_file)

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))

# Original Excel file had timestamp in different columns
input_metadata['datetime'] = input_metadata[['DATUM', 'Hour', 'Mins']].apply(lambda x: '{0} {1}-{2}'.format(datetime.strftime(x[0], '%Y-%m-%d'),"{0:0=2d}".format(x[1]),"{0:0=2d}".format(x[2])), axis = 1)
input_metadata['Detector'] = "Detector"+input_metadata['Detector'].astype('str')
result = pd.merge(input_metadata, mapping_df, how='left', on=['datetime', "Fortnight", "site", "Detector"])


#%% Map filenames to rows, verify image existence
    
start_time = time.time()
filenames_to_rows = {}
image_filenames = result[filename_col]
image_filenames = list(set(image_filenames))

missing_files = []
duplicate_rows = []

# Build up a map from filenames to a list of rows, checking image existence as we go
for iFile, fn in enumerate(image_filenames):
    try:
        if fn == 'nan' or type(fn) == float:
            pass
        else:
            if (fn in filenames_to_rows):
                duplicate_rows.append(iFile)
                filenames_to_rows[fn].append(iFile)
            else:
                filenames_to_rows[fn] = [iFile]
                image_path = os.path.join(image_directory, fn)
                if not os.path.isfile(image_path):
                    missing_files.append(fn)
    except Exception as e:
        pass

elapsed = time.time() - start_time

print('Finished verifying image existence in {}, found {} missing files (of {})'.format(
    humanfriendly.format_timespan(elapsed), 
    len(missing_files),len(image_filenames)))

#%% Skipping this check because one image has multiple species
# assert len(duplicate_rows) == 0

#%% Check for images that aren't included in the metadata file

images_missing_from_metadata = []

for iImage, image_path in enumerate(image_full_paths):

    relative_path = os.path.relpath(image_path, image_directory)
    if relative_path not in filenames_to_rows:
        images_missing_from_metadata.append(relative_path)
    
print('{} of {} files are not in metadata'.format(len(images_missing_from_metadata),len(image_full_paths)))

#%% Create CCT dictionaries

images = []
annotations = []

# Map categories to integer IDs
#
# The category '0' is reserved for 'empty'

categories_to_category_id = {}
categories_to_counts = {}
categories_to_category_id['empty'] = 0
categories_to_counts['empty'] = 0

next_category_id = 1

# For each image

start_time = time.time()
for image_name in image_filenames:

    if type(image_name) != str:
        continue

    image_path = os.path.join(image_directory, image_name)
    # Don't include images that don't exist on disk
    if not os.path.isfile(image_path):
        continue
    
    im = {}
    im['id'] = image_name.split('.')[0]
    im['file_name'] = image_name

    if load_width_and_height:
        pilImage = Image.open(image_path)
        width, height = pilImage.size
        im['width'] = width
        im['height'] = height
    else:
        im['width'] = -1
        im['height'] = -1

    images.append(im)

    rows = filenames_to_rows[image_name]

    # Some filenames will match to multiple rows
    # assert(len(rows) == 1)

    # iRow = rows[0]
    for iRow in rows:
        row = result.iloc[iRow]

        category = row['Species']

        # Have we seen this category before?
        if category in categories_to_category_id:
            categoryID = categories_to_category_id[category]
            categories_to_counts[category] += 1
        else:
            categoryID = next_category_id
            categories_to_category_id[category] = categoryID
            categories_to_counts[category] = 1
            next_category_id += 1

        # Create an annotation
        ann = {}

        # The Internet tells me this guarantees uniqueness to a reasonable extent, even
        # beyond the sheer improbability of collisions.
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']
        ann['category_id'] = categoryID
        # ann['datetime'] = row['datetime']
        # ann['site'] = row['site']

        for fieldname in annotation_fields_to_copy:
            ann[fieldname] = row[fieldname]
            if ann[fieldname] is np.nan:
                ann[fieldname] = ''
            ann[fieldname] = str(ann[fieldname])

        annotations.append(ann)

# ...for each image

# Convert categories to a CCT-style dictionary
categories = []

for category in categories_to_counts:
    print('Category {}, count {}'.format(
        category, categories_to_counts[category]))
    categoryID = categories_to_category_id[category]
    cat = {}
    cat['name'] = category
    cat['id'] = categoryID
    categories.append(cat)

elapsed = time.time() - start_time
print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))


#%% Create info struct

info = {}
info['year'] = 2012
info['version'] = 1
info['description'] = 'RSPB Dataset'
info['contributor'] = 'Helena Detection'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_json_file, 'w'), indent=4)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
    len(images), len(annotations), len(categories)))

#%% Validate output

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_directory
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False
data = integrity_check_json_db.integrity_check_json_db(output_json_file,options)


#%% Preview labels

from megadetector.visualization import visualize_db
from megadetector.data_management.databases import integrity_check_json_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = None
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['empty']
html_output_file,image_db = visualize_db.visualize_db(db_path=output_json_file,
                                                      output_dir=os.path.join(
                                                      output_directory, 'RSPB/preview'),
                                                      image_base_dir=image_directory,
                                                      options=viz_options)
os.startfile(html_output_file)




"""

 zamba_results_to_md_results.py

 Convert a labels.csv file produced by Zamba Cloud to a MD results file suitable
 for import into Timelapse.

 Columns are expected to be:

 video_uuid (not used)
 original_filename (assumed to be a relative path name)
 top_k_label,top_k_probability, for k = 1..N
 [category name 1],[category name 2],...
 corrected_label

 Because the MD results file fundamentally stores detections, what we'll
 actually do is created bogus detections that fill the entire image.  Detection
 coordinates are not currently used in Timelapse video video anyway.

 There is no special handling of empty/blank categories; because these results are
 based on a classifier, rather than a detector (where "blank" would be the absence of
 all other categories), "blank" can be queried in Timelapse just like any other class.

"""

#%% Imports and constants

import pandas as pd
import json


#%% Main function

def zamba_results_to_md_results(input_file,output_file=None):
    """
    Converts the .csv file [input_file] to the MD-formatted .json file [output_file].
    
    If [output_file] is None, '.json' will be appended to the input file.
    """
    
    if output_file is None:
        output_file = input_file + '.json'

    df = pd.read_csv(input_file)
    
    expected_columns = ('video_uuid','corrected_label','original_filename')
    for s in expected_columns:
        assert s in df.columns,\
            'Expected column {} not found, are you sure this is a Zamba results .csv file?'.format(
                s)
            
    # How many results are included per file?
    assert 'top_1_probability' in df.columns and 'top_1_label' in df.columns
    top_k = 2
    while(True):
        p_string = 'top_' + str(top_k) + '_probability'
        label_string = 'top_' + str(top_k) + '_label'
        
        if p_string in df.columns:
            assert label_string in df.columns,\
                'Oops, {} is a column but {} is not'.format(
                    p_string,label_string)
            top_k += 1
            continue
        else:
            assert label_string not in df.columns,\
                'Oops, {} is a column but {} is not'.format(
                    label_string,p_string)
            top_k -= 1
            break
            
    print('Found {} probability column pairs'.format(top_k))
    
    # Category names start after the fixed columns and the probability columns
    category_names = []
    column_names = list(df.columns)
    first_category_name_index = 0
    while('top_' in column_names[first_category_name_index] or \
          column_names[first_category_name_index] in expected_columns):
        first_category_name_index += 1
        
    i_column = first_category_name_index
    while( (i_column < len(column_names)) and (column_names[i_column] != 'corrected_label') ):
        category_names.append(column_names[i_column])
        i_column += 1
        
    print('Found {} categories:\n'.format(len(category_names)))
    
    for s in category_names:
        print(s)
    
    info = {}
    info['format_version'] = '1.3'
    info['detector'] = 'Zamba Cloud'
    info['classifier'] = 'Zamba Cloud'    
    
    detection_category_id_to_name = {}
    for category_id,category_name in enumerate(category_names):
        detection_category_id_to_name[str(category_id)] = category_name
    detection_category_name_to_id = {v: k for k, v in detection_category_id_to_name.items()}
    
    images = []
    
    # i_row = 0; row = df.iloc[i_row]
    for i_row,row in df.iterrows():
        
        im = {}
        images.append(im)
        im['file'] = row['original_filename']
        
        detections = []
        
        # k = 1
        for k in range(1,top_k+1):
            label = row['top_{}_label'.format(k)]
            confidence = row['top_{}_probability'.format(k)]
            det = {}
            det['category'] = detection_category_name_to_id[label]
            det['conf'] = confidence
            det['bbox'] = [0,0,1.0,1.0]
            detections.append(det)
            
        im['detections'] = detections
        
    # ...for each row
    
    results = {}
    results['info'] = info
    results['detection_categories'] = detection_category_id_to_name
    results['images'] = images
    
    with open(output_file,'w') as f:
        json.dump(results,f,indent=1)
        
# ...zamba_results_to_md_results(...)
        

#%% Interactive driver

if False:

    pass

    #%%
    
    input_file = r"G:\temp\labels-job-b95a4b76-e332-4e17-ab40-03469392d36a-2023-11-04_16-28-50.060130.csv"
    output_file = None
    zamba_results_to_md_results(input_file,output_file)
    

#%% Command-line driver

import sys,argparse

def main():

    parser = argparse.ArgumentParser(
        description='Convert a Zamba-formatted .csv results file to a MD-formatted .json results file')
    
    parser.add_argument(
        'input_file',
        type=str,
        help='input .csv file')
    
    parser.add_argument(
        '--output_file',
        type=str,
        default=None,
        help='output .json file (defaults to input file appended with ".json")')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    zamba_results_to_md_results(args.input_file,args.output_file)
    
if __name__ == '__main__':
    main()
    



"""

 Prepare a LILA-ready .json file for the NOAA Puget Sound Nearshore Fish dataset.

"""

#%% Constants and imports

import os
import json
import uuid
import pandas as pd

from megadetector.utils.path_utils import open_file

base_folder = r'G:\temp\noaa'
output_json_fn = os.path.join(base_folder,'noaa_estuary_fish.json')
edited_image_folders = ['edited_clip_2017','edited_clip_2018']
jpeg_image_folder = 'JPEGImages'
metadata_file = 'MasterDataForMicrosoft.xlsx'


#%% Enumerate files

edited_image_files = []

# edited_image_folder = edited_image_folders[0]
for edited_image_folder in edited_image_folders:
    folder_path = os.path.join(base_folder,edited_image_folder)
    image_files = os.listdir(folder_path)
    assert all([fn.endswith('.jpg') for fn in image_files])
    edited_image_files.extend([os.path.join(folder_path,fn) for fn in image_files])
    
jpeg_image_folder_files = os.listdir(os.path.join(base_folder,jpeg_image_folder))
assert all([fn.endswith('.jpg') for fn in jpeg_image_folder_files])

relative_edited_image_files_set = set()

# fn = edited_image_files[0]
for fn in edited_image_files:
    bn = os.path.basename(fn)
    assert bn not in relative_edited_image_files_set
    relative_edited_image_files_set.add(bn)
    
jpeg_image_folder_files_set = set(jpeg_image_folder_files)

assert len(jpeg_image_folder_files_set) == len(relative_edited_image_files_set)

assert jpeg_image_folder_files_set == relative_edited_image_files_set


#%% Read metadata and capture location information

df = pd.read_excel(os.path.join(base_folder,metadata_file))

print('Read {} rows from metadata file'.format(len(df)))

id_string_to_site = {}

# i_row = 0; row = df.iloc[i_row]
for i_row,row in df.iterrows():
    
    assert row['sd'].lower().startswith('sd')
    assert isinstance(row['id'],int) and row['id'] > 0 and row['id'] < 10000
    date_string = row['date']
    date_tokens = date_string.split('_')
    
    # Sometimes '2017' was just '17' in the date column
    if len(date_tokens[2]) != 4:
        assert len(date_tokens[2]) == 2
        date_tokens[2] = '20' + date_tokens[2]
        date_string = '_'.join(date_tokens)
    else:
        assert date_tokens[2].startswith('201')
        
    id_string = row['sd'].upper() + '_' + str(row['id']) + '_' + date_string
    id_string_to_site[id_string] = row['site']
    
print('Found {} unique locations'.format(len(pd.unique(df['site']))))


#%% Read the .json files and build output dictionaries

json_files = [fn for fn in os.listdir(base_folder) if (fn.endswith('.json') and (fn != os.path.basename(output_json_fn)))]
json_files = [os.path.join(base_folder,fn) for fn in json_files]

fn_to_image = {}
annotations = []

CATEGORY_ID_EMPTY = 0
CATEGORY_ID_FISH = 1

categories = [{'id':CATEGORY_ID_EMPTY,'name':'empty'},{'id':CATEGORY_ID_FISH,'name':'animal'}]

empty_images = set()
non_empty_images = set()

n_matched_locations = 0
images_with_unmatched_locations = []

import random
random.seed(1)

site_to_location_id = {}

# json_fn = json_files[0]
for json_fn in json_files:
    
    # if 'partial' in json_fn:
    #    continue
    
    with open(json_fn,'r') as f:
        
        lines = f.readlines()
        
        # line = lines[0]
        for line in lines:
            
            d = json.loads(line)
            image_fn = d['image']
        
            # if image_fn == 'SD1_238_6_26_17_16_76.73.jpg':
            #    asdfad
                
            # SD29_079_5_14_2018_17_52.85.jpg
            
            tokens = image_fn.split('_')
            assert len(tokens) == 7
            assert tokens[0].startswith('SD')
            
            # Re-write two-digit years as four-digit years
            if len(tokens[4]) != 4:
                assert len(tokens[4]) == 2
                tokens[4] = '20' + tokens[4]
            else:
                assert tokens[4].startswith('201')                
            
            # Sometimes the year was written with two digits instead of 4
            # assert len(tokens[4]) == 4 and tokens[4].startswith('20')
            
            while tokens[1].startswith('0'):
                tokens[1] = tokens[1][1:]
            assert not tokens[1].startswith('0')
            assert len(tokens[1]) > 0

            id_string = '_'.join(tokens[0:5])
            
            location_id = 'unknown'
                        
            if id_string in id_string_to_site:
                
                site_id = id_string_to_site[id_string]
                
                # Have we seen this location already?
                if site_id in site_to_location_id:
                    location_id = site_to_location_id[site_id]
                else:
                    location_id = 'loc_' + str(uuid.uuid1())
                    site_to_location_id[site_id] = location_id
                    print('Adding new location ID {} for site {}'.format(
                        location_id,site_id))                    
                n_matched_locations += 1
                
            else:
                raise ValueError('Could not match location ID')
                images_with_unmatched_locations.append(image_fn)
            
            assert image_fn in jpeg_image_folder_files_set
            assert d['type'] == 'image/jpg'
            input_ann = d['annotations']
            assert len(input_ann) == 1 and len(input_ann.keys()) == 1 and 'object' in input_ann
            input_ann = input_ann['object']
            assert input_ann['metainfo']['image']['height'] == 1080
            assert input_ann['metainfo']['image']['width'] == 1920
        
            im = {}
                        
            img_h = input_ann['metainfo']['image']['height']
            img_w = input_ann['metainfo']['image']['width']
            
            im['width'] = img_w
            im['height'] = img_h
            im['file_name'] = image_fn
            
            if image_fn in fn_to_image:
                assert fn_to_image[image_fn]['file_name'] == image_fn
                assert fn_to_image[image_fn]['width'] == img_w
                assert fn_to_image[image_fn]['height'] == img_h                
                im = fn_to_image[image_fn]
            else:
                fn_to_image[image_fn] = im
                im['location'] = location_id
                im['id'] = image_fn # str(uuid.uuid1())                
            
            # Not a typo, it's actually "formateddata"
            formatted_data = input_ann['formateddata']
            if len(formatted_data) == 0:
                
                # An image shouldn't be annotated as both empty and non-empty
                assert image_fn not in non_empty_images
                empty_images.add(image_fn)                
                ann = {}
                ann['id'] = str(uuid.uuid1())
                ann['image_id'] = im['id']
                ann['category_id'] = CATEGORY_ID_EMPTY
                ann['sequence_level_annotation'] = False                
                annotations.append(ann)
                
            else:
                
                # An image shouldn't be annotated as both empty and non-empty
                assert image_fn not in empty_images
                non_empty_images.add(image_fn)                
                
                n_boxes = len(formatted_data)
                
                # box = formatteddata[0]
                for box in formatted_data:
                                        
                    attributes = box['attribute']
                    assert len(attributes) == 2 and 'occluded' in attributes and 'truncated' in attributes                    
                    coordinates = box['coordinates']
                    assert box['object_type'] == 'bbox'
                    assert box['class']['type'] == 'Fish'
                    assert len(coordinates) == 4
                    for coord in coordinates:
                        assert len(coord) == 2 and 'x' in coord and 'y' in coord
                    assert coordinates[0]['y'] == coordinates[1]['y']
                    assert coordinates[2]['y'] == coordinates[3]['y']
                    assert coordinates[0]['x'] == coordinates[3]['x']
                    assert coordinates[1]['x'] == coordinates[2]['x']
                          
                    assert coordinates[0]['x'] < coordinates[1]['x']
                    assert coordinates[0]['y'] < coordinates[3]['y']
                    
                    if False:
                        x = coordinates[0]['x'] / img_w
                        y = coordinates[0]['y'] / img_h
                        box_w = (coordinates[1]['x'] - coordinates[0]['x']) / img_w
                        box_h = (coordinates[3]['y'] - coordinates[0]['y']) / img_h
                    else:
                        x = coordinates[0]['x']
                        y = coordinates[0]['y']
                        box_w = (coordinates[1]['x'] - coordinates[0]['x'])
                        box_h = (coordinates[3]['y'] - coordinates[0]['y'])
                        
                    bbox = [x,y,box_w,box_h]
                    
                    ann = {}
                    ann['id'] = str(uuid.uuid1())
                    ann['image_id'] = im['id']
                    ann['category_id'] = CATEGORY_ID_FISH
                    ann['sequence_level_annotation'] = False
                    ann['bbox'] = bbox
                    
                    annotations.append(ann)
                                        
                    # open_file(os.path.join(base_folder,jpeg_image_folder,image_fn))
                    
                # ...for each box
                
            # ...if there are boxes on this image
        
        # ...for each line
        
    # ...with open()
    
# ...for each json file
        
print('Found annotations for {} images (of {})'.format(len(fn_to_image),
                                                       len(jpeg_image_folder_files_set)))
                

print('Matched locations for {} images (failed to match {})'.format(
    n_matched_locations,len(images_with_unmatched_locations)))

images = list(fn_to_image.values())


#%% Prepare the output .json

info = {}
info['version'] = '2022.07.31.00'
info['description'] = 'NOAA Estuary Fish 2022'
info['year'] = 2022
info['contributor'] = 'NOAA Fisheries'
    
d = {}
d['info'] = info
d['annotations'] = annotations
d['images'] = images
d['categories'] = categories

with open(output_json_fn,'w') as f:
    json.dump(d,f,indent=1)


#%% Check DB integrity

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = os.path.join(base_folder,jpeg_image_folder)
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = True

_, _, _ = integrity_check_json_db.integrity_check_json_db(output_json_fn, options)


#%% Print unique locations

from collections import defaultdict
location_to_count = defaultdict(int)
for im in d['images']:
    location_to_count[im['location']] += 1
for loc in location_to_count.keys():
    print(loc + ': ' + str(location_to_count[loc]))

print('{} unique locations'.format(len(location_to_count)))
assert 'unknown' not in location_to_count.keys()

# SD12_202_6_23_2017_1_31.85.jpg


#%% Preview some images

from megadetector.visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 10000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.include_filename_links = True

html_output_file, _ = visualize_db.visualize_db(db_path=output_json_fn,
                                                output_dir=os.path.join(base_folder,'preview'),
                                                image_base_dir=os.path.join(base_folder,jpeg_image_folder),
                                                options=viz_options)
open_file(html_output_file)


#%% Statistics

print('Empty: {}'.format(len(empty_images)))
print('Non-empty: {}'.format(len(non_empty_images)))

images_with_no_boxes = 0
n_boxes = 0
for ann in annotations:
    if 'bbox' not in ann:
        images_with_no_boxes += 1
    else:
        assert len(bbox) == 4
        n_boxes += 1

print('N boxes: {}'.format(n_boxes))



"""

run_detector_batch.py

Module to run MegaDetector on lots of images, writing the results
to a file in the MegaDetector results format.

https://github.com/agentmorris/MegaDetector/tree/main/megadetector/api/batch_processing#megadetector-batch-output-format

This enables the results to be used in our post-processing pipeline; see postprocess_batch_results.py.

This script can save results to checkpoints intermittently, in case disaster
strikes. To enable this, set --checkpoint_frequency to n > 0, and results 
will be saved as a checkpoint every n images. Checkpoints will be written 
to a file in the same directory as the output_file, and after all images
are processed and final results file written to output_file, the temporary
checkpoint file will be deleted. If you want to resume from a checkpoint, set
the checkpoint file's path using --resume_from_checkpoint.

The `threshold` you can provide as an argument is the confidence threshold above
which detections will be included in the output file.

Has multiprocessing support for CPUs only; if a GPU is available, it will
use the GPU instead of CPUs, and the --ncores option will be ignored.  Checkpointing
is not supported when using a GPU.

The lack of GPU multiprocessing support might sound annoying, but in practice we
run a gazillion MegaDetector images on multiple GPUs using this script, we just only use
one GPU *per invocation of this script*.  Dividing a big batch of images into one chunk
per GPU happens outside of this script.

Does not have a command-line option to bind the process to a particular GPU, but you can 
prepend with "CUDA_VISIBLE_DEVICES=0 ", for example, to bind to GPU 0, e.g.:

CUDA_VISIBLE_DEVICES=0 python detection/run_detector_batch.py md_v4.1.0.pb ~/data ~/mdv4test.json 

You can disable GPU processing entirely by setting CUDA_VISIBLE_DEVICES=''.

"""

#%% Constants, imports, environment

import argparse
import json
import os
import sys
import time
import copy
import shutil
import warnings
import itertools
import humanfriendly

from datetime import datetime
from functools import partial
from tqdm import tqdm

import multiprocessing
from threading import Thread
from multiprocessing import Process, Manager

# Multiprocessing uses processes, not threads... leaving this here (and commented out)
# to make sure I don't change this casually at some point, it changes a number of 
# assumptions about interaction with PyTorch and TF.
# from multiprocessing.pool import ThreadPool as workerpool
from multiprocessing.pool import Pool as workerpool

from megadetector.detection import run_detector
from megadetector.detection.run_detector import \
    is_gpu_available,\
    load_detector,\
    try_download_known_detector,\
    get_detector_version_from_filename,\
    get_detector_metadata_from_version_string

from megadetector.utils import path_utils
from megadetector.visualization import visualization_utils as vis_utils
from megadetector.data_management import read_exif
from megadetector.data_management.yolo_output_to_md_output import read_classes_from_yolo_dataset_file

# Numpy FutureWarnings from tensorflow import
warnings.filterwarnings('ignore', category=FutureWarning)

# Number of images to pre-fetch
max_queue_size = 10

# How often should we print progress when using the image queue?
n_queue_print = 1000

use_threads_for_queue = False
verbose = False

exif_options = read_exif.ReadExifOptions()
exif_options.processing_library = 'pil'
exif_options.byte_handling = 'convert_to_string'


#%% Support functions for multiprocessing

def _producer_func(q,image_files):
    """ 
    Producer function; only used when using the (optional) image queue.
    
    Reads up to N images from disk and puts them on the blocking queue for processing.
    """
    
    if verbose:
        print('Producer starting'); sys.stdout.flush()
        
    for im_file in image_files:
    
        try:
            if verbose:
                print('Loading image {}'.format(im_file)); sys.stdout.flush()
            image = vis_utils.load_image(im_file)
        except Exception:
            print('Producer process: image {} cannot be loaded.'.format(im_file))
            image = run_detector.FAILURE_IMAGE_OPEN            
        
        if verbose:
            print('Queueing image {}'.format(im_file)); sys.stdout.flush()
        q.put([im_file,image])
    
    q.put(None)
        
    print('Finished image loading'); sys.stdout.flush()
    
    
def _consumer_func(q,
                   return_queue,
                   model_file,
                   confidence_threshold,
                   image_size=None,
                   include_image_size=False,
                   include_image_timestamp=False, 
                   include_exif_data=False,
                   augment=False):
    """ 
    Consumer function; only used when using the (optional) image queue.
    
    Pulls images from a blocking queue and processes them.
    """
    
    if verbose:
        print('Consumer starting'); sys.stdout.flush()

    start_time = time.time()
    detector = load_detector(model_file)
    elapsed = time.time() - start_time
    print('Loaded model (before queueing) in {}, printing updates every {} images'.format(
        humanfriendly.format_timespan(elapsed),n_queue_print))
    sys.stdout.flush()
        
    results = []
    
    n_images_processed = 0
    
    while True:
        r = q.get()
        if r is None:
            q.task_done()
            return_queue.put(results)
            return
        n_images_processed += 1
        im_file = r[0]
        image = r[1]
        if verbose or ((n_images_processed % n_queue_print) == 1):
            elapsed = time.time() - start_time
            images_per_second = n_images_processed / elapsed
            print('De-queued image {} ({:.2f}/s) ({})'.format(n_images_processed,
                                                          images_per_second,
                                                          im_file));
            sys.stdout.flush()
        if isinstance(image,str):
            # This is how the producer function communicates read errors
            results.append({'file': im_file,
                            'failure': image})
        else:
            results.append(process_image(im_file=im_file,
                                         detector=detector,
                                         confidence_threshold=confidence_threshold,
                                         image=image,
                                         quiet=True,
                                         image_size=image_size,
                                         include_image_size=include_image_size,
                                         include_image_timestamp=include_image_timestamp, 
                                         include_exif_data=include_exif_data,
                                         augment=augment))
        if verbose:
            print('Processed image {}'.format(im_file)); sys.stdout.flush()
        q.task_done()
            

def run_detector_with_image_queue(image_files,
                                  model_file,
                                  confidence_threshold,
                                  quiet=False,
                                  image_size=None,
                                  include_image_size=False, 
                                  include_image_timestamp=False,
                                  include_exif_data=False,
                                  augment=False):
    """
    Driver function for the (optional) multiprocessing-based image queue; only used 
    when --use_image_queue is specified.  Starts a reader process to read images from disk, but 
    processes images in the  process from which this function is called (i.e., does not currently
    spawn a separate consumer process).
    
    Args:
        image_files (str): list of absolute paths to images
        model_file (str): filename or model identifier (e.g. "MDV5A")
        confidence_threshold (float): minimum confidence detection to include in
            output
        quiet (bool, optional): suppress per-image console printouts
        image_size (tuple, optional): image size to use for inference, only mess with this
            if (a) you're using a model other than MegaDetector or (b) you know what you're
            doing
            
    Returns:
        list: list of dicts in the format returned by process_image()
    """
    
    q = multiprocessing.JoinableQueue(max_queue_size)
    return_queue = multiprocessing.Queue(1)
    
    if use_threads_for_queue:
        producer = Thread(target=_producer_func,args=(q,image_files,))
    else:
        producer = Process(target=_producer_func,args=(q,image_files,))
    producer.daemon = False
    producer.start()
 
    # The queue system is a little more elegant if we start one thread for reading and one
    # for processing, and this works fine on Windows, but because we import TF at module load,
    # CUDA will only work in the main process, so currently the consumer function runs here.
    #
    # To enable proper multi-GPU support, we may need to move the TF import to a separate module
    # that isn't loaded until very close to where inference actually happens.
    run_separate_consumer_process = False

    if run_separate_consumer_process:
        if use_threads_for_queue:
            consumer = Thread(target=_consumer_func,args=(q,
                                                          return_queue,
                                                          model_file,
                                                          confidence_threshold,
                                                          image_size,
                                                          include_image_size,
                                                          include_image_timestamp, 
                                                          include_exif_data,
                                                          augment))
        else:
            consumer = Process(target=_consumer_func,args=(q,
                                                           return_queue,
                                                           model_file,
                                                           confidence_threshold,
                                                           image_size,
                                                           include_image_size,
                                                           include_image_timestamp, 
                                                           include_exif_data,
                                                           augment))
        consumer.daemon = True
        consumer.start()
    else:
        _consumer_func(q,
                       return_queue,
                       model_file,
                       confidence_threshold,
                       image_size,
                       include_image_size,
                       include_image_timestamp, 
                       include_exif_data,
                       augment)

    producer.join()
    print('Producer finished')
   
    if run_separate_consumer_process:
        consumer.join()
        print('Consumer finished')
    
    q.join()
    print('Queue joined')

    results = return_queue.get()
    
    return results


#%% Other support functions

def _chunks_by_number_of_chunks(ls, n):
    """
    Splits a list into n even chunks.
    
    External callers should use ct_utils.split_list_into_n_chunks().

    Args:
        ls (list): list to break up into chunks
        n (int): number of chunks
    """
    
    for i in range(0, n):
        yield ls[i::n]


#%% Image processing functions

def process_images(im_files, 
                   detector, 
                   confidence_threshold, 
                   use_image_queue=False, 
                   quiet=False, 
                   image_size=None, 
                   checkpoint_queue=None, 
                   include_image_size=False, 
                   include_image_timestamp=False, 
                   include_exif_data=False,
                   augment=False):
    """
    Runs a detector (typically MegaDetector) over a list of image files on a single thread.
    
    Args:
        im_files (list: paths to image files                                   
        detector (str or detector object): loaded model or str; if this is a string, it can be a
            path to a .pb/.pt model file or a known model identifier (e.g. "MDV5A")
        confidence_threshold (float): only detections above this threshold are returned
        use_image_queue (bool, optional): separate image loading onto a dedicated worker process
        quiet (bool, optional): suppress per-image printouts
        image_size (tuple, optional): image size to use for inference, only mess with this
            if (a) you're using a model other than MegaDetector or (b) you know what you're
            doing
        checkpoint_queue (Queue, optional): internal parameter used to pass image queues around
        include_image_size (bool, optional): should we include image size in the output for each image?
        include_image_timestamp (bool, optional): should we include image timestamps in the output for each image?
        include_exif_data (bool, optional): should we include EXIF data in the output for each image?
        augment (bool, optional): enable image augmentation

    Returns:
        list: list of dicts, in which each dict represents detections on one image,
        see the 'images' key in https://github.com/agentmorris/MegaDetector/tree/main/megadetector/api/batch_processing#batch-processing-api-output-format
    """
    
    if isinstance(detector, str):
        
        start_time = time.time()
        detector = load_detector(detector)
        elapsed = time.time() - start_time
        print('Loaded model (batch level) in {}'.format(humanfriendly.format_timespan(elapsed)))

    if use_image_queue:
        
        run_detector_with_image_queue(im_files, 
                                      detector, 
                                      confidence_threshold, 
                                      quiet=quiet, 
                                      image_size=image_size,
                                      include_image_size=include_image_size, 
                                      include_image_timestamp=include_image_timestamp,
                                      include_exif_data=include_exif_data,
                                      augment=augment)
        
    else:            
        
        results = []
        for im_file in im_files:
            result = process_image(im_file, 
                                   detector, 
                                   confidence_threshold,
                                   quiet=quiet, 
                                   image_size=image_size, 
                                   include_image_size=include_image_size, 
                                   include_image_timestamp=include_image_timestamp,
                                   include_exif_data=include_exif_data,
                                   augment=augment)

            if checkpoint_queue is not None:
                checkpoint_queue.put(result)
            results.append(result)                                    
            
        return results

# ...def process_images(...)


def process_image(im_file, detector, 
                  confidence_threshold, 
                  image=None, 
                  quiet=False, 
                  image_size=None, 
                  include_image_size=False,
                  include_image_timestamp=False, 
                  include_exif_data=False,
                  skip_image_resizing=False,
                  augment=False):
    """
    Runs a detector (typically MegaDetector) on a single image file.

    Args:
        im_file (str): path to image file
        detector (detector object): loaded model, this can no longer be a string by the time 
            you get this far down the pipeline
        confidence_threshold (float): only detections above this threshold are returned
        image (Image, optional): previously-loaded image, if available, used when a worker
            thread is handling image loads
        quiet (bool, optional): suppress per-image printouts
        image_size (tuple, optional): image size to use for inference, only mess with this
            if (a) you're using a model other than MegaDetector or (b) you know what you're
            doing        
        include_image_size (bool, optional): should we include image size in the output for each image?
        include_image_timestamp (bool, optional): should we include image timestamps in the output for each image?
        include_exif_data (bool, optional): should we include EXIF data in the output for each image?                
        skip_image_resizing (bool, optional): whether to skip internal image resizing and rely on external resizing
        augment (bool, optional): enable image augmentation

    Returns:
        dict: dict representing detections on one image,
        see the 'images' key in 
        https://github.com/agentmorris/MegaDetector/tree/main/megadetector/api/batch_processing#batch-processing-api-output-format
    """
    
    if not quiet:
        print('Processing image {}'.format(im_file))
    
    if image is None:
        try:
            image = vis_utils.load_image(im_file)
        except Exception as e:
            if not quiet:
                print('Image {} cannot be loaded. Exception: {}'.format(im_file, e))
            result = {
                'file': im_file,
                'failure': run_detector.FAILURE_IMAGE_OPEN
            }
            return result

    try:
        result = detector.generate_detections_one_image(
                    image, 
                    im_file, 
                    detection_threshold=confidence_threshold, 
                    image_size=image_size,
                    skip_image_resizing=skip_image_resizing,
                    augment=augment)
    except Exception as e:
        if not quiet:
            print('Image {} cannot be processed. Exception: {}'.format(im_file, e))
        result = {
            'file': im_file,
            'failure': run_detector.FAILURE_INFER
        }
        return result

    if include_image_size:
        result['width'] = image.width
        result['height'] = image.height

    if include_image_timestamp:
        result['datetime'] = get_image_datetime(image)

    if include_exif_data:
        result['exif_metadata'] = read_exif.read_pil_exif(image,exif_options)

    return result

# ...def process_image(...)


def _load_custom_class_mapping(class_mapping_filename):
    """
    This is an experimental hack to allow the use of non-MD YOLOv5 models through
    the same infrastructure; it disables the code that enforces MDv5-like class lists.
    
    Should be a .json file that maps int-strings to strings, or a YOLOv5 dataset.yaml file.
    """
    
    if class_mapping_filename is None:
        return
    
    run_detector.USE_MODEL_NATIVE_CLASSES = True
    if class_mapping_filename.endswith('.json'):
        with open(class_mapping_filename,'r') as f:
            class_mapping = json.load(f)
    elif (class_mapping_filename.endswith('.yml') or class_mapping_filename.endswith('.yaml')):
        class_mapping = read_classes_from_yolo_dataset_file(class_mapping_filename)
        # convert from ints to int-strings
        class_mapping = {str(k):v for k,v in class_mapping.items()}
    else:
        raise ValueError('Unrecognized class mapping file {}'.format(class_mapping_filename))
        
    print('Loaded custom class mapping:')
    print(class_mapping)
    run_detector.DEFAULT_DETECTOR_LABEL_MAP = class_mapping
    return class_mapping
    
    
#%% Main function

def load_and_run_detector_batch(model_file, 
                                image_file_names, 
                                checkpoint_path=None,
                                confidence_threshold=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                                checkpoint_frequency=-1, 
                                results=None, 
                                n_cores=1,
                                use_image_queue=False, 
                                quiet=False, 
                                image_size=None, 
                                class_mapping_filename=None, 
                                include_image_size=False, 
                                include_image_timestamp=False, 
                                include_exif_data=False,
                                augment=False):
    """
    Load a model file and run it on a list of images.
    
    Args:
        
        model_file (str): path to model file, or supported model string (e.g. "MDV5A")
        image_file_names (list or str): list of strings (image filenames), a single image filename, 
            a folder to recursively search for images in, or a .json or .txt file containing a list 
            of images.
        checkpoint_path (str, optional), path to use for checkpoints (if None, checkpointing
            is disabled)
        confidence_threshold (float, optional): only detections above this threshold are returned
        checkpoint_frequency (int, optional): int, write results to JSON checkpoint file every N 
            images, -1 disabled checkpointing
        results (list, optional): list of dicts, existing results loaded from checkpoint; generally 
            not useful if you're using this function outside of the CLI
        n_cores (int, optional): number of parallel worker to use, ignored if we're running on a GPU
        use_image_queue (bool, optional): use a dedicated worker for image loading
        quiet (bool, optional): disable per-image console output
        image_size (tuple, optional): image size to use for inference, only mess with this
            if (a) you're using a model other than MegaDetector or (b) you know what you're
            doing
        class_mapping_filename (str, optional), use a non-default class mapping supplied in a .json 
            file or YOLOv5 dataset.yaml file
        include_image_size (bool, optional): should we include image size in the output for each image?
        include_image_timestamp (bool, optional): should we include image timestamps in the output for each image?
        include_exif_data (bool, optional): should we include EXIF data in the output for each image?
        augment (bool, optional): enable image augmentation
        
    Returns:
        results: list of dicts; each dict represents detections on one image
    """
    
    # Validate input arguments
    if n_cores is None:
        n_cores = 1
    
    if confidence_threshold is None:
        confidence_threshold=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD
    
    # Disable checkpointing if checkpoint_path is None
    if checkpoint_frequency is None or checkpoint_path is None:
        checkpoint_frequency = -1

    if class_mapping_filename is not None:
        _load_custom_class_mapping(class_mapping_filename)
        
    # Handle the case where image_file_names is not yet actually a list
    if isinstance(image_file_names,str):
        
        # Find the images to score; images can be a directory, may need to recurse
        if os.path.isdir(image_file_names):
            image_dir = image_file_names
            image_file_names = path_utils.find_images(image_dir, True)
            print('{} image files found in folder {}'.format(len(image_file_names),image_dir))
            
        # A single file, or a list of image paths
        elif os.path.isfile(image_file_names):
            list_file = image_file_names
            if image_file_names.endswith('.json'):
                with open(list_file,'r') as f:
                    image_file_names = json.load(f)
                print('Loaded {} image filenames from .json list file {}'.format(
                    len(image_file_names),list_file))
            elif image_file_names.endswith('.txt'):
                with open(list_file,'r') as f:
                    image_file_names = f.readlines()
                    image_file_names = [s.strip() for s in image_file_names if len(s.strip()) > 0]
                print('Loaded {} image filenames from .txt list file {}'.format(
                    len(image_file_names),list_file))
            elif path_utils.is_image_file(image_file_names):
                image_file_names = [image_file_names]
                print('Processing image {}'.format(image_file_names[0]))
            else:
                raise ValueError(
                    'File {} supplied as [image_file_names] argument, but extension is neither .json nor .txt'\
                        .format(
                        list_file))
        else:            
            raise ValueError(
                '{} supplied as [image_file_names] argument, but it does not appear to be a file or folder'.format(
                    image_file_names))
            
    if results is None:
        results = []

    already_processed = set([i['file'] for i in results])

    model_file = try_download_known_detector(model_file)
        
    print('GPU available: {}'.format(is_gpu_available(model_file)))
    
    if n_cores > 1 and is_gpu_available(model_file):
        
        print('Warning: multiple cores requested, but a GPU is available; parallelization across ' + \
              'GPUs is not currently supported, defaulting to one GPU')
        n_cores = 1

    if n_cores > 1 and use_image_queue:
        
        print('Warning: multiple cores requested, but the image queue is enabled; parallelization ' + \
              'with the image queue is not currently supported, defaulting to one worker')
        n_cores = 1
        
    if use_image_queue:
        
        assert checkpoint_frequency < 0, \
            'Using an image queue is not currently supported when checkpointing is enabled'
        assert len(results) == 0, \
            'Using an image queue with results loaded from a checkpoint is not currently supported'
        assert n_cores <= 1
        results = run_detector_with_image_queue(image_file_names, 
                                                model_file, 
                                                confidence_threshold, 
                                                quiet, 
                                                image_size=image_size,
                                                include_image_size=include_image_size,
                                                include_image_timestamp=include_image_timestamp,
                                                include_exif_data=include_exif_data,
                                                augment=augment)
        
    elif n_cores <= 1:

        # Load the detector
        start_time = time.time()
        detector = load_detector(model_file)
        elapsed = time.time() - start_time
        print('Loaded model in {}'.format(humanfriendly.format_timespan(elapsed)))

        # This is only used for console reporting, so it's OK that it doesn't
        # include images we might have loaded from a previous checkpoint
        count = 0

        for im_file in tqdm(image_file_names):

            # Will not add additional entries not in the starter checkpoint
            if im_file in already_processed:
                if not quiet:
                    print('Bypassing image {}'.format(im_file))
                continue

            count += 1

            result = process_image(im_file, 
                                   detector, 
                                   confidence_threshold, 
                                   quiet=quiet, 
                                   image_size=image_size, 
                                   include_image_size=include_image_size,
                                   include_image_timestamp=include_image_timestamp,
                                   include_exif_data=include_exif_data,
                                   augment=augment)
            results.append(result)

            # Write a checkpoint if necessary
            if (checkpoint_frequency != -1) and ((count % checkpoint_frequency) == 0):
                
                print('Writing a new checkpoint after having processed {} images since '
                      'last restart'.format(count))
                
                _write_checkpoint(checkpoint_path, results)
            
    else:
        
        # Multiprocessing is enabled at this point
        
        # When using multiprocessing, tell the workers to load the model on each
        # process, by passing the model_file string as the "model" argument to
        # process_images.
        detector = model_file

        print('Creating pool with {} cores'.format(n_cores))

        if len(already_processed) > 0:
            n_images_all = len(image_file_names)
            image_file_names = [fn for fn in image_file_names if fn not in already_processed]
            print('Loaded {} of {} images from checkpoint'.format(
                len(already_processed),n_images_all))
        
        # Divide images into chunks; we'll send one chunk to each worker process   
        image_batches = list(_chunks_by_number_of_chunks(image_file_names, n_cores))
                
        pool = workerpool(n_cores)

        if checkpoint_path is not None:
            
            # Multiprocessing and checkpointing are both enabled at this point
            
            checkpoint_queue = Manager().Queue()
            
            # Pass the "results" array (which may already contain images loaded from an existing
            # checkpoint) to the checkpoint queue handler function, which will append results to 
            # the list as they become available.
            checkpoint_thread = Thread(target=_checkpoint_queue_handler, 
                                       args=(checkpoint_path, checkpoint_frequency,
                                             checkpoint_queue, results), daemon=True)
            checkpoint_thread.start()

            pool.map(partial(process_images, 
                             detector=detector,
                             confidence_threshold=confidence_threshold,
                             use_image_queue=False,
                             quiet=quiet,
                             image_size=image_size, 
                             checkpoint_queue=checkpoint_queue,
                             include_image_size=include_image_size,
                             include_image_timestamp=include_image_timestamp,
                             include_exif_data=include_exif_data,
                             augment=augment), 
                             image_batches)

            checkpoint_queue.put(None)

        else:
            
            # Multprocessing is enabled, but checkpointing is not
            
            new_results = pool.map(partial(process_images, 
                                           detector=detector,
                                           confidence_threshold=confidence_threshold,
                                           use_image_queue=False,
                                           quiet=quiet,
                                           checkpoint_queue=None,
                                           image_size=image_size,
                                           include_image_size=include_image_size,
                                           include_image_timestamp=include_image_timestamp,
                                           include_exif_data=include_exif_data,
                                           augment=augment), 
                                           image_batches)

            new_results = list(itertools.chain.from_iterable(new_results))
            
            # Append the results we just computed to "results", which is *usually* empty, but will
            # be non-empty if we resumed from a checkpoint
            results += new_results

        # ...if checkpointing is/isn't enabled
        
    # ...if we're running (1) with image queue, (2) on one core, (3) on multiple cores
    
    # 'results' may have been modified in place, but we also return it for
    # backwards-compatibility.
    return results

# ...def load_and_run_detector_batch(...)


def _checkpoint_queue_handler(checkpoint_path, checkpoint_frequency, checkpoint_queue, results):
    """
    Thread function to accumulate results and write checkpoints when checkpointing and
    multiprocessing are both enabled.
    """
    
    result_count = 0
    while True:
        result = checkpoint_queue.get()        
        if result is None:            
            break  
        
        result_count +=1
        results.append(result)

        if (checkpoint_frequency != -1) and (result_count % checkpoint_frequency == 0):
                
            print('Writing a new checkpoint after having processed {} images since '
                    'last restart'.format(result_count))
            
            _write_checkpoint(checkpoint_path, results)


def _write_checkpoint(checkpoint_path, results):
    """
    Writes the 'images' field in the dict 'results' to a json checkpoint file.
    """
    
    assert checkpoint_path is not None             
            
    # Back up any previous checkpoints, to protect against crashes while we're writing
    # the checkpoint file.
    checkpoint_tmp_path = None
    if os.path.isfile(checkpoint_path):
        checkpoint_tmp_path = checkpoint_path + '_tmp'
        shutil.copyfile(checkpoint_path,checkpoint_tmp_path)
        
    # Write the new checkpoint
    with open(checkpoint_path, 'w') as f:
        json.dump({'images': results}, f, indent=1, default=str)
        
    # Remove the backup checkpoint if it exists
    if checkpoint_tmp_path is not None:
        os.remove(checkpoint_tmp_path)


def get_image_datetime(image):
    """
    Reads EXIF datetime from a PIL Image object.
    
    Args:
        image (Image): the PIL Image object from which we should read datetime information
        
    Returns:
        str: the EXIF datetime from [image] (a PIL Image object), if available, as a string;
        returns None if EXIF datetime is not available.
    """
    
    exif_tags = read_exif.read_pil_exif(image,exif_options)
    
    try:
        datetime_str = exif_tags['DateTimeOriginal']
        _ = time.strptime(datetime_str, '%Y:%m:%d %H:%M:%S')
        return datetime_str

    except Exception:
        return None        


def write_results_to_file(results, 
                          output_file, 
                          relative_path_base=None, 
                          detector_file=None, 
                          info=None, 
                          include_max_conf=False,
                          custom_metadata=None, 
                          force_forward_slashes=True):
    """
    Writes list of detection results to JSON output file. Format matches:

    https://github.com/agentmorris/MegaDetector/tree/main/megadetector/api/batch_processing#batch-processing-api-output-format

    Args:
        results (list):  list of dict, each dict represents detections on one image
        output_file (str): path to JSON output file, should end in '.json'
        relative_path_base (str, optional): path to a directory as the base for relative paths, can
            be None if the paths in [results] are absolute
        detector_file (str, optional): filename of the detector used to generate these results, only
            used to pull out a version number for the "info" field
        info (dict, optional): dictionary to put in the results file instead of the default "info" field
        include_max_conf (bool, optional): old files (version 1.2 and earlier) included a "max_conf" field
            in each image; this was removed in version 1.3.  Set this flag to force the inclusion
            of this field.
        custom_metadata (object, optional): additional data to include as info['custom_metadata']; typically
            a dictionary, but no type/format checks are performed
        force_forward_slashes (bool, optional): convert all slashes in filenames within [results] to
            forward slashes
                    
    Returns:
        dict: the MD-formatted dictionary that was written to [output_file]
    """
    
    if relative_path_base is not None:
        results_relative = []
        for r in results:
            r_relative = copy.copy(r)
            r_relative['file'] = os.path.relpath(r_relative['file'], start=relative_path_base)
            results_relative.append(r_relative)
        results = results_relative

    if force_forward_slashes:
        results_converted = []
        for r in results:
            r_converted = copy.copy(r)
            r_converted['file'] = r_converted['file'].replace('\\','/')
            results_converted.append(r_converted)
        results = results_converted
            
    # The typical case: we need to build the 'info' struct
    if info is None:
        
        info = { 
            'detection_completion_time': datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'),
            'format_version': '1.3' 
        }
        
        if detector_file is not None:
            detector_filename = os.path.basename(detector_file)
            detector_version = get_detector_version_from_filename(detector_filename)
            detector_metadata = get_detector_metadata_from_version_string(detector_version)
            info['detector'] = detector_filename  
            info['detector_metadata'] = detector_metadata
        else:
            info['detector'] = 'unknown'
            info['detector_metadata'] = get_detector_metadata_from_version_string('unknown')
        
    # If the caller supplied the entire "info" struct
    else:
        
        if detector_file is not None:
            
            print('Warning (write_results_to_file): info struct and detector file ' + \
                  'supplied, ignoring detector file')

    if custom_metadata is not None:
        info['custom_metadata'] = custom_metadata
        
    # The 'max_detection_conf' field used to be included by default, and it caused all kinds
    # of headaches, so it's no longer included unless the user explicitly requests it.
    if not include_max_conf:
        for im in results:
            if 'max_detection_conf' in im:
                del im['max_detection_conf']
            
    final_output = {
        'images': results,
        'detection_categories': run_detector.DEFAULT_DETECTOR_LABEL_MAP,
        'info': info
    }
    
    with open(output_file, 'w') as f:
        json.dump(final_output, f, indent=1, default=str)
    print('Output file saved at {}'.format(output_file))
    
    return final_output

# ...def write_results_to_file(...)


#%% Interactive driver

if False:
    
    pass

    #%%
    
    model_file = 'MDV5A'
    image_dir = r'g:\camera_traps\camera_trap_images'
    output_file = r'g:\temp\md-test.json'
    
    recursive = True
    output_relative_filenames = True
    include_max_conf = False
    quiet = True
    image_size = None
    use_image_queue = False
    confidence_threshold = 0.0001
    checkpoint_frequency = 5   
    checkpoint_path = None
    resume_from_checkpoint = 'auto'
    allow_checkpoint_overwrite = False
    ncores = 1
    class_mapping_filename = None
    include_image_size = True
    include_image_timestamp = True
    include_exif_data = True
    overwrite_handling = None
        
    # Generate a command line
    cmd = 'python run_detector_batch.py "{}" "{}" "{}"'.format(
        model_file,image_dir,output_file)
    
    if recursive:
        cmd += ' --recursive'
    if output_relative_filenames:
        cmd += ' --output_relative_filenames'
    if include_max_conf:
        cmd += ' --include_max_conf'
    if quiet:
        cmd += ' --quiet'
    if image_size is not None:
        cmd += ' --image_size {}'.format(image_size)
    if use_image_queue:
        cmd += ' --use_image_queue'
    if confidence_threshold is not None:
        cmd += ' --threshold {}'.format(confidence_threshold)
    if checkpoint_frequency is not None:
        cmd += ' --checkpoint_frequency {}'.format(checkpoint_frequency)
    if checkpoint_path is not None:
        cmd += ' --checkpoint_path "{}"'.format(checkpoint_path)
    if resume_from_checkpoint is not None:
        cmd += ' --resume_from_checkpoint "{}"'.format(resume_from_checkpoint)
    if allow_checkpoint_overwrite:
        cmd += ' --allow_checkpoint_overwrite'
    if ncores is not None:
        cmd += ' --ncores {}'.format(ncores)
    if class_mapping_filename is not None:
        cmd += ' --class_mapping_filename "{}"'.format(class_mapping_filename)
    if include_image_size:
        cmd += ' --include_image_size'
    if include_image_timestamp:
        cmd += ' --include_image_timestamp'
    if include_exif_data:
        cmd += ' --include_exif_data'
    if overwrite_handling is not None:
        cmd += ' --overwrite_handling {}'.format(overwrite_handling)
    
    print(cmd)
    import clipboard; clipboard.copy(cmd)
    
    
    #%% Run inference interactively
    
    image_file_names = path_utils.find_images(image_dir, recursive=False)    
    results = None
    
    start_time = time.time()
    
    results = load_and_run_detector_batch(model_file=model_file,
                                          image_file_names=image_file_names,
                                          checkpoint_path=checkpoint_path,
                                          confidence_threshold=confidence_threshold,
                                          checkpoint_frequency=checkpoint_frequency,
                                          results=results,
                                          n_cores=ncores,
                                          use_image_queue=use_image_queue,
                                          quiet=quiet,
                                          image_size=image_size)
    
    elapsed = time.time() - start_time
    
    print('Finished inference in {}'.format(humanfriendly.format_timespan(elapsed)))

    
#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser(
        description='Module to run a TF/PT animal detection model on lots of images')
    parser.add_argument(
        'detector_file',
        help='Path to detector model file (.pb or .pt).  Can also be the strings "MDV4", "MDV5A", or "MDV5B" to request automatic download.')
    parser.add_argument(
        'image_file',
        help=\
        'Path to a single image file, a .json or .txt file containing a list of paths to images, or a directory')
    parser.add_argument(
        'output_file',
        help='Path to output JSON results file, should end with a .json extension')
    parser.add_argument(
        '--recursive',
        action='store_true',
        help='Recurse into directories, only meaningful if image_file points to a directory')
    parser.add_argument(
        '--output_relative_filenames',
        action='store_true',
        help='Output relative file names, only meaningful if image_file points to a directory')
    parser.add_argument(
        '--include_max_conf',
        action='store_true',
        help='Include the "max_detection_conf" field in the output')
    parser.add_argument(
        '--quiet',
        action='store_true',
        help='Suppress per-image console output')
    parser.add_argument(
        '--image_size',
        type=int,
        default=None,
        help=('Force image resizing to a specific integer size on the long axis (not recommended to change this)'))    
    parser.add_argument(
        '--augment',
        action='store_true',
        help='Enable image augmentation'
    )
    parser.add_argument(
        '--use_image_queue',
        action='store_true',
        help='Pre-load images, may help keep your GPU busy; does not currently support ' + \
             'checkpointing.  Useful if you have a very fast GPU and a very slow disk.')
    parser.add_argument(
        '--threshold',
        type=float,
        default=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
        help="Confidence threshold between 0 and 1.0, don't include boxes below this " + \
            "confidence in the output file. Default is {}".format(
                run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD))
    parser.add_argument(
        '--checkpoint_frequency',
        type=int,
        default=-1,
        help='Write results to a temporary file every N images; default is -1, which ' + \
             'disables this feature')
    parser.add_argument(
        '--checkpoint_path',
        type=str,
        default=None,
        help='File name to which checkpoints will be written if checkpoint_frequency is > 0, ' + \
             'defaults to md_checkpoint_[date].json in the same folder as the output file')    
    parser.add_argument(
        '--resume_from_checkpoint',
        type=str,
        default=None,
        help='Path to a JSON checkpoint file to resume from, or "auto" to ' + \
             'find the most recent checkpoint in the same folder as the output file.  "auto" uses' + \
             'checkpoint_path (rather than searching the output folder) if checkpoint_path is specified.')
    parser.add_argument(
        '--allow_checkpoint_overwrite',
        action='store_true',
        help='By default, this script will bail if the specified checkpoint file ' + \
              'already exists; this option allows it to overwrite existing checkpoints')
    parser.add_argument(
        '--ncores',
        type=int,
        default=0,
        help='Number of cores to use; only applies to CPU-based inference')
    parser.add_argument(
        '--class_mapping_filename',
        type=str,
        default=None,
        help='Use a non-default class mapping, supplied in a .json file with a dictionary mapping' + \
            'int-strings to strings.  This will also disable the addition of "1" to all category ' + \
            'IDs, so your class mapping should start at zero.  Can also be a YOLOv5 dataset.yaml file.')
    parser.add_argument(
        '--include_image_size',
        action='store_true',
        help='Include image dimensions in output file'
    )
    parser.add_argument(
        '--include_image_timestamp',
        action='store_true',
        help='Include image datetime (if available) in output file'
    )
    parser.add_argument(
        '--include_exif_data',
        action='store_true',
        help='Include available EXIF data in output file'
    )
    parser.add_argument(
        '--overwrite_handling',
        type=str,
        default='overwrite',
        help='What should we do if the output file exists?  overwrite/skip/error (default overwrite)'
    )
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # If the specified detector file is really the name of a known model, find 
    # (and possibly download) that model
    args.detector_file = try_download_known_detector(args.detector_file)
    
    assert os.path.exists(args.detector_file), \
        'detector file {} does not exist'.format(args.detector_file)
    assert 0.0 <= args.threshold <= 1.0, 'Confidence threshold needs to be between 0 and 1'
    assert args.output_file.endswith('.json'), 'output_file specified needs to end with .json'
    if args.checkpoint_frequency != -1:
        assert args.checkpoint_frequency > 0, 'Checkpoint_frequency needs to be > 0 or == -1'
    if args.output_relative_filenames:
        assert os.path.isdir(args.image_file), \
            f'Could not find folder {args.image_file}, must supply a folder when ' + \
                '--output_relative_filenames is set'

    if os.path.exists(args.output_file):
        if args.overwrite_handling == 'overwrite':
            print('Warning: output file {} already exists and will be overwritten'.format(
                args.output_file))
        elif args.overwrite_handling == 'skip':
            print('Output file {} exists, returning'.format(
                args.output_file))
            return
        elif args.overwrite_handling == 'error':
            raise Exception('Output file {} exists'.format(args.output_file))
        else:
            raise ValueError('Illegal overwrite handling string {}'.format(args.overwrite_handling))

    output_dir = os.path.dirname(args.output_file)

    if len(output_dir) > 0:
        os.makedirs(output_dir,exist_ok=True)
        
    assert not os.path.isdir(args.output_file), 'Specified output file is a directory'
    
    if args.class_mapping_filename is not None:
        _load_custom_class_mapping(args.class_mapping_filename)
    
    # Load the checkpoint if available
    #
    # Relative file names are only output at the end; all file paths in the checkpoint are
    # still absolute paths.
    if args.resume_from_checkpoint is not None:
        if args.resume_from_checkpoint == 'auto':
            checkpoint_files = os.listdir(output_dir)
            checkpoint_files = [fn for fn in checkpoint_files if \
                                (fn.startswith('md_checkpoint') and fn.endswith('.json'))]
            if len(checkpoint_files) == 0:
                raise ValueError('resume_from_checkpoint set to "auto", but no checkpoints found in {}'.format(
                    output_dir))
            else:
                if len(checkpoint_files) > 1:
                    print('Warning: found {} checkpoints in {}, using the latest'.format(
                        len(checkpoint_files),output_dir))
                    checkpoint_files = sorted(checkpoint_files)
                checkpoint_file_relative = checkpoint_files[-1]
                checkpoint_file = os.path.join(output_dir,checkpoint_file_relative)                                
        else:
            checkpoint_file = args.resume_from_checkpoint
        assert os.path.exists(checkpoint_file), \
            'File at resume_from_checkpoint specified does not exist'
        with open(checkpoint_file) as f:
            print('Loading previous results from checkpoint file {}'.format(
                checkpoint_file))
            saved = json.load(f)
        assert 'images' in saved, \
            'The checkpoint file does not have the correct fields; cannot be restored'
        results = saved['images']
        print('Restored {} entries from the checkpoint'.format(len(results)))
    else:
        results = []

    # Find the images to score; images can be a directory, may need to recurse
    if os.path.isdir(args.image_file):
        image_file_names = path_utils.find_images(args.image_file, args.recursive)
        if len(image_file_names) > 0:
            print('{} image files found in the input directory'.format(len(image_file_names)))                        
        else:
            if (args.recursive):
                print('No image files found in directory {}, exiting'.format(args.image_file))
            else:
                print('No image files found in directory {}, did you mean to specify '
                      '--recursive?'.format(
                    args.image_file))
            return
        
    # A json list of image paths
    elif os.path.isfile(args.image_file) and args.image_file.endswith('.json'):
        with open(args.image_file) as f:
            image_file_names = json.load(f)
        print('Loaded {} image filenames from .json list file {}'.format(
            len(image_file_names),args.image_file))
    
    # A text list of image paths
    elif os.path.isfile(args.image_file) and args.image_file.endswith('.txt'):
        with open(args.image_file) as f:
            image_file_names = f.readlines()
            image_file_names = [fn.strip() for fn in image_file_names if len(fn.strip()) > 0]
        print('Loaded {} image filenames from .txt list file {}'.format(
            len(image_file_names),args.image_file))
        
    # A single image file
    elif os.path.isfile(args.image_file) and path_utils.is_image_file(args.image_file):
        image_file_names = [args.image_file]
        print('Processing image {}'.format(args.image_file))
        
    else:        
        raise ValueError('image_file specified is not a directory, a json list, or an image file, '
                         '(or does not have recognizable extensions).')

    assert len(image_file_names) > 0, 'Specified image_file does not point to valid image files'
    assert os.path.exists(image_file_names[0]), \
        'The first image to be processed does not exist at {}'.format(image_file_names[0])

    # Test that we can write to the output_file's dir if checkpointing requested
    if args.checkpoint_frequency != -1:
        
        if args.checkpoint_path is not None:
            checkpoint_path = args.checkpoint_path
        else:
            checkpoint_path = os.path.join(output_dir,
                                           'md_checkpoint_{}.json'.format(
                                               datetime.utcnow().strftime("%Y%m%d%H%M%S")))
        
        # Don't overwrite existing checkpoint files, this is a sure-fire way to eventually
        # erase someone's checkpoint.
        if (checkpoint_path is not None) and (not args.allow_checkpoint_overwrite) \
            and (args.resume_from_checkpoint is None):
            
            assert not os.path.isfile(checkpoint_path), \
                f'Checkpoint path {checkpoint_path} already exists, delete or move it before ' + \
                're-using the same checkpoint path, or specify --allow_checkpoint_overwrite'

        
        # Confirm that we can write to the checkpoint path; this avoids issues where
        # we crash after several thousand images.
        #
        # But actually, commenting this out for now... the scenario where we are resuming from a 
        # checkpoint, then immediately overwrite that checkpoint with empty data is higher-risk
        # than the annoyance of crashing a few minutes after starting a job.
        if False:
            with open(checkpoint_path, 'w') as f:
                json.dump({'images': []}, f)
                
        print('The checkpoint file will be written to {}'.format(checkpoint_path))
        
    else:
        
        if args.checkpoint_path is not None:
            print('Warning: checkpointing disabled because checkpoint_frequency is -1, ' + \
                  'but a checkpoint path was specified')
        checkpoint_path = None

    start_time = time.time()

    results = load_and_run_detector_batch(model_file=args.detector_file,
                                          image_file_names=image_file_names,
                                          checkpoint_path=checkpoint_path,
                                          confidence_threshold=args.threshold,
                                          checkpoint_frequency=args.checkpoint_frequency,
                                          results=results,
                                          n_cores=args.ncores,
                                          use_image_queue=args.use_image_queue,
                                          quiet=args.quiet,
                                          image_size=args.image_size,
                                          class_mapping_filename=args.class_mapping_filename,
                                          include_image_size=args.include_image_size,
                                          include_image_timestamp=args.include_image_timestamp,
                                          include_exif_data=args.include_exif_data,
                                          augment=args.augment)

    elapsed = time.time() - start_time
    images_per_second = len(results) / elapsed
    print('Finished inference for {} images in {} ({:.2f} images per second)'.format(
        len(results),humanfriendly.format_timespan(elapsed),images_per_second))

    relative_path_base = None
    if args.output_relative_filenames:
        relative_path_base = args.image_file
    write_results_to_file(results, args.output_file, relative_path_base=relative_path_base,
                          detector_file=args.detector_file,include_max_conf=args.include_max_conf)

    if checkpoint_path and os.path.isfile(checkpoint_path):
        os.remove(checkpoint_path)
        print('Deleted checkpoint file {}'.format(checkpoint_path))

    print('Done, thanks for MegaDetect\'ing!')

if __name__ == '__main__':
    main()



"""

 rspb_to_json.py

 Convert the .csv file provided for the RSPB data set to a 
 COCO-camera-traps .json file

"""

#%% Constants and environment

import pandas as pd
import os
import glob
import json
import re
import uuid
import tqdm
import time
import ntpath
import humanfriendly
import PIL

from megadetector.data_management.databases import integrity_check_json_db
from megadetector.visualization import visualize_db

# [location] is an obfuscation
baseDir = r'e:\wildlife_data\rspb_gola_data'
metadataFile = os.path.join(baseDir,'gola_camtrapr_master_renaming_table_2019-01-31.csv')
outputFile = os.path.join(baseDir,'rspb_gola_labeled.json')
imageBaseDir = os.path.join(baseDir,'gola_camtrapr_data')
imageFlatDir = os.path.join(baseDir,'gola_camtrapr_data_flat')
unmatchedImagesFile = os.path.join(baseDir,'unmatchedImages.txt')
assert(os.path.isdir(imageBaseDir))


#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'COCO style database for RSPB gola data'
info['secondary_contributor'] = 'Converted to COCO .json by Dan Morris'
info['contributor'] = 'RSPB'


#%% Read source data

metadataTable = pd.read_csv(metadataFile)

print('Read {} columns and {} rows from metadata file'.format(len(metadataTable.columns),
      len(metadataTable)))

# metadataTable.columns.values
#
# array(['Project', 'inDir', 'FileName', 'Station', 'Camera',
#        'StationCameraFileName', 'DateTimeOriginal', 'DateReadable',
#        'outDir', 'filename_new', 'fileExistsAlready', 'CopyStatus',
#        'Species'], dtype=object)

metadataTable[['Species']] = metadataTable[['Species']].fillna(value='unlabeled')

# We'll populate these later 
metadataTable['sequenceID'] = ''
metadataTable['frameNumber'] = ''
metadataTable['filePath'] = ''

failedCopies = metadataTable[~metadataTable.CopyStatus]
print('Removing {} rows that were failed copies'.format(len(failedCopies)))

metadataTable = metadataTable[metadataTable.CopyStatus]

species = list(metadataTable.Species)
uniqueSpecies = set(species)

print('Read {} unique species in {} rows'.format(len(uniqueSpecies),len(metadataTable)))

speciesMappings = {}

# keys should be lowercase
speciesMappings['blank'] = 'empty'
speciesMappings[''] = 'unlabeled'


#%% Enumerate images, confirm filename uniqueness

imageFullPaths = glob.glob(os.path.join(imageBaseDir,r'**\*.JPG'),recursive=True)

print('Counted {} images'.format(len(imageFullPaths)))

filenamesOnly = set()

for p in imageFullPaths:
    
    fn = ntpath.basename(p)
    assert fn not in filenamesOnly
    filenamesOnly.add(fn)
    
print('Finished uniqueness checking')


#%% Update metadata filenames to include site and camera folders, check existence
#
# Takes ~1min

filenamesToRows = {}

startTime = time.time()

newRows = []
matchFailures = []

# iRow = 0; row = metadataTable.iloc[iRow]
for iRow,row in tqdm.tqdm(metadataTable.iterrows(), total=metadataTable.shape[0]):
    
    baseFn = row['filename_new']
    station = row['Station']
    
    filenamesToRows[baseFn] = iRow
    
    # There's a bug in the metadata; the 'camera' column isn't correct.  
    # camera = row['Camera']
    # These appear as, e.g., '3.22e12'    
    # camera = str(int(float(camera)))
    
    # Let's pull this out of the file name instead
    #
    # Filenames look like one of the following:
    #
    # A1__03224850850507__2015-11-28__10-45-04(1).JPG
    # Bayama2PH__C05__NA(NA).JPG
    pat = '^(?P<station>.+?)__(?P<camera>.+?)__((?P<date>.+?)__)?(?P<time>[^_\()]+?)\((?P<frame>.+?)\)\.JPG'
    match = re.match(pat,baseFn)
    if match is None:
        raise ValueError('Regex failure at row {}: {}'.format(iRow,baseFn))
    assert(station == match.group('station'))
    camera = match.group('camera')
    row['Camera'] = camera
    
    assert match.group('station') is not None
    assert match.group('camera') is not None
    assert match.group('frame') is not None
    
    if match.group('date') is None:
        imgDate = ''
    else:
        imgDate = match.group('date')
        
    if match.group('time') is None:
        imgTime = ''
    else:
        imgTime = match.group('time')
    
    frame = -1
    try:
        frame = int(match.group['frame'])
    except:
        pass
    row['frameNumber'] = frame
    
    fn = os.path.join(station,camera,baseFn)
    fullPath = os.path.join(imageBaseDir,fn)
    row['filePath'] = fn
    # assert(os.path.isfile(fullPath))
    if not os.path.isfile(fullPath):
        print('Failed to match image {}'.format(fullPath))
        matchFailures.append(fullPath)
        continue
    
    # metadataTable.iloc[iRow] = row
    newRows.append(row)

elapsed = time.time() - startTime

# Re-assemble into an updated table
metadataTable = pd.DataFrame(newRows)

print('Finished checking file existence, extracting metadata in {}, couldn''t find {} images'.format(
      humanfriendly.format_timespan(elapsed),len(matchFailures)))
       
    
#%% Check for images that aren't included in the metadata file

imagesNotInMetadata = []

# Enumerate all images
for iImage,imagePath in enumerate(imageFullPaths):
    
    fn = ntpath.basename(imagePath)
    if(fn not in filenamesToRows):
        imagesNotInMetadata.append(imagePath)

print('Finished matching {} images, failed to match {}'.format(
        len(imageFullPaths),len(imagesNotInMetadata)))

# Write to a text file
with open(unmatchedImagesFile, 'w') as f:
    for fn in imagesNotInMetadata:
        f.write('{}\n'.format(fn))
        
        
#%% Create CCT dictionaries

# Also gets image sizes, so this takes ~6 minutes
#
# Implicitly checks images for overt corruptness, i.e. by not crashing.

images = []
annotations = []

# Map categories to integer IDs (that's what COCO likes)
nextCategoryID = 1
categoriesToCategoryId = {'empty':0}
categoriesToCounts = {'empty':0}

# For each image
#
# Because in practice images are 1:1 with annotations in this data set,
# this is also a loop over annotations.

startTime = time.time()

# iRow = 0; row = metadataTable.iloc[iRow]
for iRow,row in tqdm.tqdm(metadataTable.iterrows(), total=metadataTable.shape[0]):
    
    im = {}
        
    # A1__03224850850507__2015-11-28__10-45-04(1).JPG
    fn = row['filename_new']
    assert '.JPG' in fn
    fn = fn.replace('.JPG','')
    im['id'] = fn
    
    # 'A1\\03224850850507\\A1__03224850850507__2015-11-28__10-45-04(1).JPG'
    im['file_name'] = row['filePath']
    
    # Not currently populated
    im['seq_id'] = row['sequenceID']
    
    # Often -1, sometimes a semi-meaningful int
    im['frame_num'] = row['frameNumber']
    
    # A1
    im['site']= row['Station']    
    
    # 03224850850507
    im['camera'] = row['Camera']
    
    # In variable form, but sometimes '28/11/2015 10:45'
    im['datetime'] = row['DateTimeOriginal']
    
    images.append(im)
    
    # Check image height and width
    imagePath = os.path.join(imageBaseDir,im['file_name'])
    assert(os.path.isfile(imagePath))
    pilImage = PIL.Image.open(imagePath)
    width, height = pilImage.size
    im['width'] = width
    im['height'] = height

    category = row['Species'].lower()
    if category in speciesMappings:
        category = speciesMappings[category]
    
    # Have we seen this category before?
    if category in categoriesToCategoryId:
        categoryID = categoriesToCategoryId[category]
        categoriesToCounts[category] += 1
    else:
        categoryID = nextCategoryID
        categoriesToCategoryId[category] = categoryID
        categoriesToCounts[category] = 0
        nextCategoryID += 1
    
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = categoryID
    
    annotations.append(ann)
    
# ...for each image
    
# Convert categories to a CCT-style dictionary

categories = []

for category in categoriesToCounts:
    
    print('Category {}, count {}'.format(category,categoriesToCounts[category]))
    categoryID = categoriesToCategoryId[category]
    cat = {}
    cat['name'] = category
    cat['id'] = categoryID
    categories.append(cat)    
    
elapsed = time.time() - startTime

print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))
    

#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data,open(outputFile,'w'),indent=4)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Check database integrity

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = imageBaseDir
options.bCheckImageSizes = False
options.bFindUnusedImages = False
integrity_check_json_db.integrity_check_json_db(outputFile, options)


#%% Preview a few images to make sure labels were passed along sensibly

db_path = outputFile
output_dir = os.path.join(baseDir,'label_preview')
image_base_dir = imageBaseDir
options = visualize_db.DbVizOptions()
options.num_to_visualize = 100
htmlOutputFile = visualize_db.visualize_db(db_path,output_dir,image_base_dir,options)
    

#%% One-time processing step: copy images to a flat directory for annotation

if False:
    
    #%%
    
    from shutil import copyfile
    os.makedirs(imageFlatDir,exist_ok=True)
    
    for sourcePath in tqdm.tqdm(imageFullPaths):
        fn = ntpath.basename(sourcePath)
        targetPath = os.path.join(imageFlatDir,fn)
        assert not os.path.isfile(targetPath)
        copyfile(sourcePath,targetPath)
        
    print('Copied {} files'.format(len(imageFullPaths)))
    


"""

azure_utils.py

Miscellaneous Azure Blob Storage utilities

Requires azure-storage-blob>=12.4.0

"""

#%% Imports

import json

from typing import Any, Iterable, List, Optional, Tuple, Union
from azure.storage.blob import BlobPrefix, ContainerClient

from megadetector.utils import path_utils
from megadetector.utils import sas_blob_utils


#%% Functions

def walk_container(container_client: ContainerClient,
                   max_depth: int = -1,
                   prefix: str = '',
                   store_folders: bool = True,
                   store_blobs: bool = True,
                   debug_max_items: int = -1) -> Tuple[List[str], List[str]]:
    """
    Recursively walk folders a Azure Blob Storage container.

    Based on:
    https://github.com/Azure/azure-sdk-for-python/blob/master/sdk/storage/azure-storage-blob/samples/blob_samples_walk_blob_hierarchy.py
    """
    
    depth = 1

    def walk_blob_hierarchy(prefix: str,
                            folders: Optional[List[str]] = None,
                            blobs: Optional[List[str]] = None
                            ) -> Tuple[List[str], List[str]]:
        if folders is None:
            folders = []
        if blobs is None:
            blobs = []

        nonlocal depth

        if 0 < max_depth < depth:
            return folders, blobs

        for item in container_client.walk_blobs(name_starts_with=prefix):
            short_name = item.name[len(prefix):]
            if isinstance(item, BlobPrefix):
                # print('F: ' + prefix + short_name)
                if store_folders:
                    folders.append(prefix + short_name)
                depth += 1
                walk_blob_hierarchy(item.name, folders=folders, blobs=blobs)
                if (debug_max_items > 0
                        and len(folders) + len(blobs) > debug_max_items):
                    return folders, blobs
                depth -= 1
            else:
                if store_blobs:
                    blobs.append(prefix + short_name)

        return folders, blobs

    folders, blobs = walk_blob_hierarchy(prefix=prefix)

    assert all(s.endswith('/') for s in folders)
    folders = [s.strip('/') for s in folders]

    return folders, blobs


def list_top_level_blob_folders(container_client: ContainerClient) -> List[str]:
    """
    List all top-level folders in a container.
    """
    
    top_level_folders, _ = walk_container(
        container_client, max_depth=1, store_blobs=False)
    return top_level_folders


def concatenate_json_lists(input_files: Iterable[str],
                           output_file: Optional[str] = None
                           ) -> List[Any]:
    """
    Given a list of JSON files that contain lists (typically string
    filenames), concatenates the lists into a single list and optionally
    writes out this list to a new output JSON file.
    """
    
    output_list = []
    for fn in input_files:
        with open(fn, 'r') as f:
            file_list = json.load(f)
        output_list.extend(file_list)
    if output_file is not None:
        with open(output_file, 'w') as f:
            json.dump(output_list, f, indent=1)
    return output_list


def upload_file_to_blob(account_name: str,
                        container_name: str,
                        local_path: str,
                        blob_name: str,
                        sas_token: str,
                        overwrite: bool=False) -> str:
    """
    Uploads a local file to Azure Blob Storage and returns the uploaded
    blob URI with SAS token.
    """
    
    container_uri = sas_blob_utils.build_azure_storage_uri(
        account=account_name, container=container_name, sas_token=sas_token)
    with open(local_path, 'rb') as data:
        return sas_blob_utils.upload_blob(
            container_uri=container_uri, blob_name=blob_name, data=data, 
            overwrite=overwrite)


def enumerate_blobs_to_file(
        output_file: str,
        account_name: str,
        container_name: str,
        sas_token: Optional[str] = None,
        blob_prefix: Optional[str] = None,
        blob_suffix: Optional[Union[str, Tuple[str]]] = None,
        rsearch: Optional[str] = None,
        limit: Optional[int] = None,
        verbose: Optional[bool] = True
        ) -> List[str]:
    """
    Enumerates blobs in a container, and writes the blob names to an output
    file.

    Args:
        output_file: str, path to save list of files in container
            If ends in '.json', writes a JSON string. Otherwise, writes a
            newline-delimited list. Can be None, in which case this is just a 
            convenient wrapper for blob enumeration.
        account_name: str, Azure Storage account name
        container_name: str, Azure Blob Storage container name
        sas_token: optional str, container SAS token, leading ? will be removed if present.
        blob_prefix: optional str, returned results will only contain blob names
            to with this prefix
        blob_suffix: optional str or tuple of str, returned results will only
            contain blob names with this/these suffix(es). The blob names will
            be lowercased first before comparing with the suffix(es).
        rsearch: optional str, returned results will only contain blob names
            that match this regex. Can also be a list of regexes, in which case
            blobs matching *any* of the regex's will be returned.            
        limit: int, maximum # of blob names to list
            if None, then returns all blob names

    Returns: list of str, sorted blob names, of length limit or shorter.
    """
    
    if sas_token is not None and len(sas_token) > 9 and sas_token[0] == '?':
        sas_token = sas_token[1:]
        
    container_uri = sas_blob_utils.build_azure_storage_uri(
        account=account_name, container=container_name, sas_token=sas_token)
    
    matched_blobs = sas_blob_utils.list_blobs_in_container(
        container_uri=container_uri, blob_prefix=blob_prefix,
        blob_suffix=blob_suffix, rsearch=rsearch, limit=limit, verbose=verbose)
    
    if output_file is not None:
        path_utils.write_list_to_file(output_file, matched_blobs)
        
    return matched_blobs
